# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for the :mod:`pennylane` :class:`QNode` class.
"""
import contextlib
import io
import math
import textwrap

import pytest
import numpy as np
from scipy.linalg import block_diag

from pennylane.plugins.default_qubit import Y, Z

import pennylane as qml
from pennylane._device import Device
from pennylane.qnode import QNode, QuantumFunctionError


class TestQNodeOperationQueue:
    """Tests that the QNode operation queue is properly filled and interacted with"""

    @pytest.fixture(scope="function")
    def qnode(self, mock_device):
        """Provides a circuit for the subsequent tests of the operation queue"""

        def circuit(x):
            qml.RX(x, wires=[0])
            qml.CNOT(wires=[0, 1])
            qml.RY(0.4, wires=[0])
            qml.RZ(-0.2, wires=[1])
            return qml.expval(qml.PauliX(0)), qml.expval(qml.PauliZ(1))

        node = qml.QNode(circuit, mock_device)
        node.construct([1.0])

        return node

    def test_operation_ordering(self, qnode):
        """Tests that the ordering of the operations is correct"""

        assert qnode.ops[0].name == "RX"
        assert qnode.ops[1].name == "CNOT"
        assert qnode.ops[2].name == "RY"
        assert qnode.ops[3].name == "RZ"
        assert qnode.ops[4].name == "PauliX"
        assert qnode.ops[5].name == "PauliZ"

    def test_op_successors_operations_only(self, qnode):
        """Tests that _op_successors properly extracts the successors that are operations"""

        operation_successors = qnode._op_successors(qnode.ops[0], only="G")

        assert qnode.ops[0] not in operation_successors
        assert qnode.ops[1] in operation_successors
        assert qnode.ops[4] not in operation_successors

    def test_op_successors_observables_only(self, qnode):
        """Tests that _op_successors properly extracts the successors that are observables"""

        observable_successors = qnode._op_successors(qnode.ops[0], only="E")

        assert qnode.ops[0] not in observable_successors
        assert qnode.ops[1] not in observable_successors
        assert qnode.ops[4] in observable_successors

    def test_op_successors_both_operations_and_observables(self, qnode):
        """Tests that _op_successors properly extracts all successors"""

        successors = qnode._op_successors(qnode.ops[0], only=None)

        assert qnode.ops[0] not in successors
        assert qnode.ops[1] in successors
        assert qnode.ops[4] in successors

    def test_op_successors_both_operations_and_observables_nodes(self, qnode):
        """Tests that _op_successors properly extracts all successor nodes"""

        successors = qnode._op_successors(qnode.ops[0], only=None)

        assert qnode.circuit.operations[0] not in successors
        assert qnode.circuit.operations[1] in successors
        assert qnode.circuit.operations[2] in successors
        assert qnode.circuit.operations[3] in successors
        assert qnode.circuit.observables[0] in successors

    def test_op_successors_both_operations_and_observables_strict_ordering(self, qnode):
        """Tests that _op_successors properly extracts all successors"""

        successors = qnode._op_successors(qnode.ops[2], only=None)

        assert qnode.circuit.operations[0] not in successors
        assert qnode.circuit.operations[1] not in successors
        assert qnode.circuit.operations[2] not in successors
        assert qnode.circuit.operations[3] not in successors
        assert qnode.circuit.observables[0] in successors

    def test_op_successors_extracts_all_successors(self, qnode):
        """Tests that _op_successors properly extracts all successors"""
        successors = qnode._op_successors(qnode.ops[2], only=None)
        assert qnode.ops[4] in successors
        assert qnode.ops[5] not in successors

    def test_print_applied(self, mock_device):
        """Test that printing applied gates works correctly"""

        H = np.array([[0, 1], [1, 0]])

        def circuit(x):
            qml.RX(x, wires=[0])
            qml.CNOT(wires=[0, 1])
            qml.RY(0.4, wires=[0])
            qml.RZ(-0.2, wires=[1])
            return qml.expval(qml.PauliX(0)), qml.var(qml.Hermitian(H, wires=1))

        expected_qnode_print = textwrap.dedent("""\
            Operations
            ==========
            RX({x}, wires=[0])
            CNOT(wires=[0, 1])
            RY(0.4, wires=[0])
            RZ(-0.2, wires=[1])

            Observables
            ===========
            expval(PauliX(wires=[0]))
            var(Hermitian([[0 1]
             [1 0]], wires=[1]))""")

        node = qml.QNode(circuit, mock_device)

        # test before construction
        f = io.StringIO()

        with contextlib.redirect_stdout(f):
            node.print_applied()
            out = f.getvalue().strip()

        assert out == "QNode has not yet been executed."

        # construct QNode
        f = io.StringIO()
        node.construct([0.1])

        with contextlib.redirect_stdout(f):
            node.print_applied()
            out = f.getvalue().strip()

        assert out == expected_qnode_print.format(x=0.1)


@pytest.fixture(scope="function")
def operable_mock_device_2_wires(monkeypatch):
    """A mock instance of the abstract Device class that can support qfuncs."""

    dev = Device
    with monkeypatch.context() as m:
        m.setattr(dev, '__abstractmethods__', frozenset())
        m.setattr(dev, 'operations', ["RX", "RY", "CNOT"])
        m.setattr(dev, 'observables', ["PauliX", "PauliY", "PauliZ"])
        m.setattr(dev, 'reset', lambda self: None)
        m.setattr(dev, 'apply', lambda self, x, y, z: None)
        m.setattr(dev, 'expval', lambda self, x, y, z: 1)
        yield Device(wires=2)


class TestQNodeBestMethod:
    """
    Test different flows of _best_method
    """
    def test_best_method_with_non_gaussian_successors(self, tol, gaussian_device_2_wires):
        """Tests that the analytic differentiation method is allowed and matches numerical
        differentiation if a non-Gaussian gate is not succeeded by an observable."""

        @qml.qnode(gaussian_device_2_wires)
        def circuit(x):
            qml.Squeezing(x, 0, wires=[0])
            qml.Beamsplitter(np.pi/4, 0, wires=[0, 1])
            qml.Kerr(0.54, wires=[1])
            return qml.expval(qml.NumberOperator(0))

        res = circuit.jacobian([0.321], method='A')
        expected = circuit.jacobian([0.321], method='F')
        assert np.allclose(res, expected, atol=tol, rtol=0)

    def test_best_method_with_gaussian_successors_fails(self, gaussian_device_2_wires):
        """Tests that the analytic differentiation method is not allowed
        if a non-Gaussian gate is succeeded by an observable."""

        @qml.qnode(gaussian_device_2_wires)
        def circuit(x):
            qml.Squeezing(x, 0, wires=[0])
            qml.Beamsplitter(np.pi/4, 0, wires=[0, 1])
            qml.Kerr(0.54, wires=[1])
            return qml.expval(qml.NumberOperator(1))

        with pytest.raises(ValueError, match="analytic gradient method cannot be used with"):
            circuit.jacobian([0.321], method='A')


class TestQNodeExceptions:
    """Tests that QNode raises proper errors"""

    def test_current_context_modified_outside_construct(self, mock_device, monkeypatch):
        """Tests that the QNode properly raises an error if the _current_context
           was modified outside of construct"""

        def circuit(x):
            qml.RX(x, wires=[0])
            return qml.expval(qml.PauliZ(wires=0))

        node = qml.QNode(circuit, mock_device)

        monkeypatch.setattr(QNode, "_current_context", node)

        with pytest.raises(
            QuantumFunctionError,
            match="QNode._current_context must not be modified outside this method.",
        ):
            node.construct([0.0])

    def test_return_of_non_observable(self, operable_mock_device_2_wires):
        """Tests that the QNode properly raises an error if the qfunc returns something
           besides observables."""

        def circuit(x):
            qml.RX(x, wires=[0])
            return qml.expval(qml.PauliZ(wires=0)), 0.3

        node = qml.QNode(circuit, operable_mock_device_2_wires)

        with pytest.raises(QuantumFunctionError, match="must return either"):
            node(0.5)

    def test_observable_not_returned(self, operable_mock_device_2_wires):
        """Tests that the QNode properly raises an error if the qfunc does not
           return all observables."""

        def circuit(x):
            qml.RX(x, wires=[0])
            ex = qml.expval(qml.PauliZ(wires=1))
            return qml.expval(qml.PauliZ(wires=0))

        node = qml.QNode(circuit, operable_mock_device_2_wires)

        with pytest.raises(QuantumFunctionError, match="All measured observables"):
            node(0.5)

    def test_observable_order_violated(self, operable_mock_device_2_wires):
        """Tests that the QNode properly raises an error if the qfunc does not
           return all observables in the correct order."""

        def circuit(x):
            qml.RX(x, wires=[0])
            ex = qml.expval(qml.PauliZ(wires=1))
            return qml.expval(qml.PauliZ(wires=0)), ex

        node = qml.QNode(circuit, operable_mock_device_2_wires)

        with pytest.raises(QuantumFunctionError, match="All measured observables"):
            node(0.5)

    def test_operations_after_observables(self, operable_mock_device_2_wires):
        """Tests that the QNode properly raises an error if the qfunc contains
           operations after observables."""

        def circuit(x):
            qml.RX(x, wires=[0])
            ex = qml.expval(qml.PauliZ(wires=1))
            qml.RY(0.5, wires=[0])
            return qml.expval(qml.PauliZ(wires=0))

        node = qml.QNode(circuit, operable_mock_device_2_wires)

        with pytest.raises(QuantumFunctionError, match="gates must precede"):
            node(0.5)

    def test_multiple_measurements_on_same_wire(self, operable_mock_device_2_wires):
        """Tests that the QNode properly raises an error if the same wire
           is measured multiple times."""

        def circuit(x):
            qml.RX(x, wires=[0])
            qml.CNOT(wires=[0, 1])
            return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliZ(1)), qml.expval(qml.PauliX(0))

        node = qml.QNode(circuit, operable_mock_device_2_wires)

        with pytest.raises(QuantumFunctionError, match="can only be measured once"):
            node(0.5)

    def test_operation_on_nonexistant_wire(self, operable_mock_device_2_wires):
        """Tests that the QNode properly raises an error if an operation
           is applied to a non-existant wire."""

        operable_mock_device_2_wires.num_wires = 2

        def circuit(x):
            qml.RX(x, wires=[0])
            qml.CNOT(wires=[0, 2])
            return qml.expval(qml.PauliZ(0))

        node = qml.QNode(circuit, operable_mock_device_2_wires)

        with pytest.raises(QuantumFunctionError, match="applied to invalid wire"):
            node(0.5)

    def test_observable_on_nonexistant_wire(self, operable_mock_device_2_wires):
        """Tests that the QNode properly raises an error if an observable
           is measured on a non-existant wire."""

        operable_mock_device_2_wires.num_wires = 2

        def circuit(x):
            qml.RX(x, wires=[0])
            return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliZ(2))

        node = qml.QNode(circuit, operable_mock_device_2_wires)

        with pytest.raises(QuantumFunctionError, match="applied to invalid wire"):
            node(0.5)

    def test_mixing_of_cv_and_qubit_operations(self, operable_mock_device_2_wires):
        """Tests that the QNode properly raises an error if qubit and
           CV operations are mixed in the same qfunc."""

        def circuit(x):
            qml.RX(x, wires=[0])
            qml.Displacement(0.5, 0, wires=[0])
            return qml.expval(qml.PauliZ(0))

        node = qml.QNode(circuit, operable_mock_device_2_wires)

        with pytest.raises(QuantumFunctionError, match="Continuous and discrete"):
            node(0.5)

    def test_transform_observable_incorrect_heisenberg_size(self):
        """Test that an exception is raised in the case that the
        dimensions of a CV observable Heisenberg representation does not match
        the ev_order attribute"""

        dev = qml.device("default.gaussian", wires=1)

        class P(qml.operation.CVObservable):
            """Dummy CV observable with incorrect ev_order"""
            num_wires = 1
            num_params = 0
            par_domain = None
            ev_order = 2

            @staticmethod
            def _heisenberg_rep(p):
                return np.array([0, 1, 0])

        def circuit(x):
            qml.Displacement(x, 0.1, wires=0)
            return qml.expval(P(0))

        node = qml.QNode(circuit, dev)

        with pytest.raises(QuantumFunctionError, match="Mismatch between polynomial order"):
            node.jacobian([0.5])


class TestQNodeJacobianExceptions:
    """Tests that QNode.jacobian raises proper errors"""

    def test_undifferentiable_operation(self, operable_mock_device_2_wires):
        """Tests that QNode.jacobian properly raises an error if the
           qfunc contains an operation that is not differentiable."""

        def circuit(x):
            qml.BasisState(np.array([x, 0]), wires=[0, 1])
            qml.RX(x, wires=[0])
            return qml.expval(qml.PauliZ(0))

        node = qml.QNode(circuit, operable_mock_device_2_wires)

        with pytest.raises(ValueError, match="Cannot differentiate wrt parameter"):
            node.jacobian(0.5)

    def test_operation_not_supporting_analytic_gradient(self, operable_mock_device_2_wires):
        """Tests that QNode.jacobian properly raises an error if the
           qfunc contains an operation that does not support analytic gradients."""

        def circuit(x):
            qml.RX(x, wires=[0])
            return qml.expval(qml.Hermitian(np.diag([x, 0]), 0))

        node = qml.QNode(circuit, operable_mock_device_2_wires)

        with pytest.raises(ValueError, match="analytic gradient method cannot be used with"):
            node.jacobian(0.5, method="A")

    def test_bogus_gradient_method_set(self, operable_mock_device_2_wires):
        """Tests that QNode.jacobian properly raises an error if the
           gradient method set is bogus."""

        def circuit(x):
            qml.RX(x, wires=[0])
            return qml.expval(qml.PauliZ(0))

        # in non-cached mode, the grad method would be
        # recomputed and overwritten from the
        # bogus value 'J'. Caching stops this from happening.
        node = qml.QNode(circuit, operable_mock_device_2_wires, cache=True)

        node.evaluate([0.0])
        keys = node.grad_method_for_par.keys()
        if keys:
            k0 = [k for k in keys][0]

        node.grad_method_for_par[k0] = "J"

        with pytest.raises(ValueError, match="Unknown gradient method"):
            node.jacobian(0.5)

    def test_indices_not_unique(self, operable_mock_device_2_wires):
        """Tests that QNode.jacobian properly raises an error if the
           jacobian is requested for non-unique indices."""

        def circuit(x):
            qml.Rot(0.3, x, -0.2, wires=[0])
            return qml.expval(qml.PauliZ(0))

        node = qml.QNode(circuit, operable_mock_device_2_wires)

        with pytest.raises(ValueError, match="Parameter indices must be unique."):
            node.jacobian(0.5, which=[0, 0])

    def test_indices_nonexistant(self, operable_mock_device_2_wires):
        """Tests that QNode.jacobian properly raises an error if the
           jacobian is requested for non-existant parameters."""

        def circuit(x):
            qml.Rot(0.3, x, -0.2, wires=[0])
            return qml.expval(qml.PauliZ(0))

        node = qml.QNode(circuit, operable_mock_device_2_wires)

        with pytest.raises(ValueError, match="Tried to compute the gradient wrt"):
            node.jacobian(0.5, which=[0, 6])

        with pytest.raises(ValueError, match="Tried to compute the gradient wrt"):
            node.jacobian(0.5, which=[1, -1])

    def test_unknown_method(self, operable_mock_device_2_wires):
        """Tests that QNode.jacobian properly raises an error if the
           gradient method is unknown."""

        def circuit(x):
            qml.Rot(0.3, x, -0.2, wires=[0])
            return qml.expval(qml.PauliZ(0))

        node = qml.QNode(circuit, operable_mock_device_2_wires)

        with pytest.raises(ValueError, match="Unknown gradient method"):
            node.jacobian(0.5, method="unknown")

    def test_wrong_order_in_finite_difference(self, operable_mock_device_2_wires):
        """Tests that QNode.jacobian properly raises an error if finite
           differences are attempted with wrong order."""

        def circuit(x):
            qml.Rot(0.3, x, -0.2, wires=[0])
            return qml.expval(qml.PauliZ(0))

        node = qml.QNode(circuit, operable_mock_device_2_wires)

        with pytest.raises(ValueError, match="Order must be 1 or 2"):
            node.jacobian(0.5, method="F", order=3)


class TestQNodeParameters:
    """Tests the handling of parameters in the QNode"""

    @pytest.mark.parametrize(
        "x,y",
        zip(np.linspace(-2 * np.pi, 2 * np.pi, 7), np.linspace(-2 * np.pi, 2 * np.pi, 7) ** 2 / 11),
    )
    def test_fanout(self, qubit_device_1_wire, tol, x, y):
        """Tests that qnodes can compute the correct function when the
           same parameter is used in multiple gates."""

        def circuit(x, y):
            qml.RX(x, wires=[0])
            qml.RZ(y, wires=[0])
            qml.RX(x, wires=[0])
            return qml.expval(qml.PauliZ(0))

        def analytic_expval(x, y):
            return math.cos(x) ** 2 - math.cos(y) * math.sin(x) ** 2

        node = qml.QNode(circuit, qubit_device_1_wire)

        assert np.isclose(node(x, y), analytic_expval(x, y), atol=tol, rtol=0)

    def test_array_parameters_scalar_return(self, qubit_device_1_wire, tol):
        """Test that QNode can take arrays as input arguments, and that they interact properly with Autograd.
           Test case for a circuit that returns a scalar."""

        def circuit(dummy1, array, dummy2):
            qml.RY(0.5 * array[0, 1], wires=0)
            qml.RY(-0.5 * array[1, 1], wires=0)
            return qml.expval(qml.PauliX(0))

        node = qml.QNode(circuit, qubit_device_1_wire)

        args = (0.46, np.array([[2.0, 3.0, 0.3], [7.0, 4.0, 2.1]]), -0.13)
        grad_target = (
            np.array(1.0),
            np.array([[0.5, 0.43879, 0], [0, -0.43879, 0]]),
            np.array(-0.4),
        )
        cost_target = 1.03257

        def cost(x, array, y):
            c = node(0.111, array, 4.5)
            return c + 0.5 * array[0, 0] + x - 0.4 * y

        cost_grad = qml.grad(cost, argnum=[0, 1, 2])
        computed_grad = cost_grad(*args)

        assert np.isclose(cost(*args), cost_target, atol=tol, rtol=0)

        assert np.allclose(computed_grad[0], grad_target[0], atol=tol, rtol=0)
        assert np.allclose(computed_grad[1], grad_target[1], atol=tol, rtol=0)
        assert np.allclose(computed_grad[2], grad_target[2], atol=tol, rtol=0)

    def test_qnode_array_parameters_1_vector_return(self, qubit_device_1_wire, tol):
        """Test that QNode can take arrays as input arguments, and that they interact properly with Autograd.
           Test case for a circuit that returns a 1-vector."""

        def circuit(dummy1, array, dummy2):
            qml.RY(0.5 * array[0, 1], wires=0)
            qml.RY(-0.5 * array[1, 1], wires=0)
            return (qml.expval(qml.PauliX(0)),)

        node = qml.QNode(circuit, qubit_device_1_wire)

        args = (0.46, np.array([[2.0, 3.0, 0.3], [7.0, 4.0, 2.1]]), -0.13)
        grad_target = (
            np.array(1.0),
            np.array([[0.5, 0.43879, 0], [0, -0.43879, 0]]),
            np.array(-0.4),
        )
        cost_target = 1.03257

        def cost(x, array, y):
            c = node(0.111, array, 4.5)[0]
            return c + 0.5 * array[0, 0] + x - 0.4 * y

        cost_grad = qml.grad(cost, argnum=[0, 1, 2])
        computed_grad = cost_grad(*args)

        assert np.isclose(cost(*args), cost_target, atol=tol, rtol=0)

        assert np.allclose(computed_grad[0], grad_target[0], atol=tol, rtol=0)
        assert np.allclose(computed_grad[1], grad_target[1], atol=tol, rtol=0)
        assert np.allclose(computed_grad[2], grad_target[2], atol=tol, rtol=0)

    def test_qnode_array_parameters_2_vector_return(self, qubit_device_2_wires, tol):
        """Test that QNode can take arrays as input arguments, and that they interact properly with Autograd.
           Test case for a circuit that returns a 2-vector."""

        def circuit(dummy1, array, dummy2):
            qml.RY(0.5 * array[0, 1], wires=0)
            qml.RY(-0.5 * array[1, 1], wires=0)
            qml.RY(array[1, 0], wires=1)
            return qml.expval(qml.PauliX(0)), qml.expval(qml.PauliX(1))

        node = qml.QNode(circuit, qubit_device_2_wires)

        args = (0.46, np.array([[2.0, 3.0, 0.3], [7.0, 4.0, 2.1]]), -0.13)
        grad_target = (
            np.array(1.0),
            np.array([[0.5, 0.43879, 0], [0, -0.43879, 0]]),
            np.array(-0.4),
        )
        cost_target = 1.03257

        def cost(x, array, y):
            c = node(0.111, array, 4.5)[0]
            return c + 0.5 * array[0, 0] + x - 0.4 * y

        cost_grad = qml.grad(cost, argnum=[0, 1, 2])
        computed_grad = cost_grad(*args)

        assert np.isclose(cost(*args), cost_target, atol=tol, rtol=0)

        assert np.allclose(computed_grad[0], grad_target[0], atol=tol, rtol=0)
        assert np.allclose(computed_grad[1], grad_target[1], atol=tol, rtol=0)
        assert np.allclose(computed_grad[2], grad_target[2], atol=tol, rtol=0)

    def test_array_parameters_evaluate(self, qubit_device_2_wires, tol):
        """Tests that array parameters gives same result as positional arguments."""
        a, b, c = 0.5, 0.54, 0.3

        def ansatz(x, y, z):
            qml.QubitStateVector(np.array([1, 0, 1, 1]) / np.sqrt(3), wires=[0, 1])
            qml.Rot(x, y, z, wires=0)
            qml.CNOT(wires=[0, 1])
            return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliY(1))

        @qml.qnode(qubit_device_2_wires)
        def circuit1(x, y, z):
            return ansatz(x, y, z)

        @qml.qnode(qubit_device_2_wires)
        def circuit2(x, array):
            return ansatz(x, array[0], array[1])

        @qml.qnode(qubit_device_2_wires)
        def circuit3(array):
            return ansatz(*array)

        positional_res = circuit1(a, b, c)
        positional_grad = circuit1.jacobian([a, b, c])

        array_res = circuit2(a, np.array([b, c]))
        array_grad = circuit2.jacobian([a, np.array([b, c])])

        assert np.allclose(positional_res, array_res, atol=tol, rtol=0)
        assert np.allclose(positional_grad, array_grad, atol=tol, rtol=0)

        list_res = circuit2(a, [b, c])
        list_grad = circuit2.jacobian([a, [b, c]])

        assert np.allclose(positional_res, list_res, atol=tol, rtol=0)
        assert np.allclose(positional_grad, list_grad, atol=tol, rtol=0)

        array_res = circuit3(np.array([a, b, c]))
        array_grad = circuit3.jacobian([np.array([a, b, c])])

        list_res = circuit3([a, b, c])
        list_grad = circuit3.jacobian([[a, b, c]])

        assert np.allclose(positional_res, array_res, atol=tol, rtol=0)
        assert np.allclose(positional_grad, array_grad, atol=tol, rtol=0)

    def test_multiple_expectation_different_wires(self, qubit_device_2_wires, tol):
        """Tests that qnodes return multiple expectation values."""

        a, b, c = 0.5, 0.54, 0.3

        @qml.qnode(qubit_device_2_wires)
        def circuit(x, y, z):
            qml.RX(x, wires=[0])
            qml.RZ(y, wires=[0])
            qml.CNOT(wires=[0, 1])
            qml.RY(y, wires=[0])
            qml.RX(z, wires=[0])
            return qml.expval(qml.PauliY(0)), qml.expval(qml.PauliZ(1))

        def analytic_expval(a, b, c):
            return [-1 * math.cos(a) * math.cos(b) * math.sin(c), math.cos(a)]

        res = circuit(a, b, c)
        analytic_res = analytic_expval(a, b, c)

        assert np.allclose(res, analytic_res, atol=tol, rtol=0)


class TestQNodeKeywordArguments:
    """Tests that the qnode properly handles keyword arguments."""

    def test_multiple_keywordargs_used(self, qubit_device_2_wires, tol):
        """Tests that qnodes use multiple keyword arguments."""

        def circuit(w, x=None, y=None):
            qml.RX(x, wires=[0])
            qml.RX(y, wires=[1])
            return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliZ(1))

        node = qml.QNode(circuit, qubit_device_2_wires)

        c = node(1.0, x=np.pi, y=np.pi)

        assert np.allclose(c, [-1.0, -1.0], atol=tol, rtol=0)

    def test_multidimensional_keywordargs_used(self, qubit_device_2_wires, tol):
        """Tests that qnodes use multi-dimensional keyword arguments."""

        def circuit(w, x=None):
            qml.RX(x[0], wires=[0])
            qml.RX(x[1], wires=[1])
            return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliZ(1))

        node = qml.QNode(circuit, qubit_device_2_wires)

        c = node(1.0, x=[np.pi, np.pi])

        assert np.allclose(c, [-1.0, -1.0], atol=tol, rtol=0)

    def test_keywordargs_for_wires(self, qubit_device_2_wires, tol):
        """Tests that wires can be passed as keyword arguments."""

        default_q = 0

        def circuit(x, q=default_q):
            qml.RX(x, wires=[q])
            return qml.expval(qml.PauliZ(q))

        node = qml.QNode(circuit, qubit_device_2_wires)

        c = node(np.pi, q=1)

        assert node.queue[0].wires == [1]
        assert np.isclose(c, -1.0, atol=tol, rtol=0)

        c = node(np.pi)

        assert node.queue[0].wires == [default_q]
        assert np.isclose(c, -1.0, atol=tol, rtol=0)

    def test_keywordargs_used(self, qubit_device_1_wire, tol):
        """Tests that qnodes use keyword arguments."""

        def circuit(w, x=None):
            qml.RX(x, wires=[0])
            return qml.expval(qml.PauliZ(0))

        node = qml.QNode(circuit, qubit_device_1_wire)

        c = node(1.0, x=np.pi)

        assert np.isclose(c, -1.0, atol=tol, rtol=0)

    def test_keywordarg_updated_in_multiple_calls(self, qubit_device_2_wires, tol):
        """Tests that qnodes update keyword arguments in consecutive calls."""

        def circuit(w, x=None):
            qml.RX(w, wires=[0])
            qml.RX(x, wires=[1])
            return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliZ(1))

        node = qml.QNode(circuit, qubit_device_2_wires)

        c1 = node(0.1, x=0.0)
        c2 = node(0.1, x=np.pi)

        assert c1[1] != c2[1]

    def test_keywordarg_passes_through_classicalnode(self, qubit_device_2_wires, tol):
        """Tests that qnodes' keyword arguments pass through classical nodes."""

        def circuit(w, x=None):
            qml.RX(w, wires=[0])
            qml.RX(x, wires=[1])
            return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliZ(1))

        node = qml.QNode(circuit, qubit_device_2_wires)

        def classical_node(w, x=None):
            return node(w, x=x)

        c = classical_node(0.0, x=np.pi)

        assert np.allclose(c, [1.0, -1.0], atol=tol, rtol=0)


class TestQNodeGradients:
    """Qnode gradient tests."""

    @pytest.mark.parametrize("shape", [(8,), (8, 1), (4, 2), (2, 2, 2), (2, 1, 2, 1, 2)])
    def test_multidim_array(self, shape, tol):
        """Tests that arguments which are multidimensional arrays are
        properly evaluated and differentiated in QNodes."""

        base_array = np.linspace(-1.0, 1.0, 8)
        multidim_array = np.reshape(base_array, shape)

        def circuit(w):
            qml.RX(w[np.unravel_index(0, shape)], wires=0)  # base_array[0]
            qml.RX(w[np.unravel_index(1, shape)], wires=1)  # base_array[1]
            qml.RX(w[np.unravel_index(2, shape)], wires=2)  # ...
            qml.RX(w[np.unravel_index(3, shape)], wires=3)
            qml.RX(w[np.unravel_index(4, shape)], wires=4)
            qml.RX(w[np.unravel_index(5, shape)], wires=5)
            qml.RX(w[np.unravel_index(6, shape)], wires=6)
            qml.RX(w[np.unravel_index(7, shape)], wires=7)
            return tuple(qml.expval(qml.PauliZ(idx)) for idx in range(len(base_array)))

        dev = qml.device("default.qubit", wires=8)
        circuit = qml.QNode(circuit, dev)

        # circuit evaluations
        circuit_output = circuit(multidim_array)
        expected_output = np.cos(base_array)
        assert np.allclose(circuit_output, expected_output, atol=tol, rtol=0)

        # circuit jacobians
        circuit_jacobian = circuit.jacobian([multidim_array])
        expected_jacobian = -np.diag(np.sin(base_array))
        assert np.allclose(circuit_jacobian, expected_jacobian, atol=tol, rtol=0)

    def test_qnode_cv_gradient_methods(self):
        """Tests the gradient computation methods on CV circuits."""
        # we can only use the 'A' method on parameters which only affect gaussian operations
        # that are not succeeded by nongaussian operations

        par = [0.4, -2.3]
        dev = qml.device("default.qubit", wires=2)

        def check_methods(qf, d):
            q = qml.QNode(qf, dev)
            # NOTE: the default plugin is a discrete (qubit) simulator, it cannot
            # execute CV gates, but the QNode can be constructed
            q.construct(par)
            assert q.grad_method_for_par == d

        def qf(x, y):
            qml.Displacement(x, 0, wires=[0])
            qml.CubicPhase(0.2, wires=[0])
            qml.Squeezing(0.3, y, wires=[1])
            qml.Rotation(1.3, wires=[1])
            # nongaussian succeeding x but not y
            qml.Kerr(0.4, wires=[0])
            return qml.expval(qml.X(0)), qml.expval(qml.X(1))

        check_methods(qf, {0: "F", 1: "A"})

        def qf(x, y):
            qml.Displacement(x, 0, wires=[0])
            qml.CubicPhase(0.2, wires=[0])  # nongaussian succeeding x
            qml.Squeezing(0.3, x, wires=[1])  # x affects gates on both wires, y unused
            qml.Rotation(1.3, wires=[1])
            return qml.expval(qml.X(0)), qml.expval(qml.X(1))

        check_methods(qf, {0: "F"})

        def qf(x, y):
            qml.Displacement(x, 0, wires=[0])
            qml.Displacement(1.2, y, wires=[0])
            qml.Beamsplitter(0.2, 1.7, wires=[0, 1])
            qml.Rotation(1.9, wires=[0])
            qml.Kerr(0.3, wires=[1])  # nongaussian succeeding both x and y due to the beamsplitter
            return qml.expval(qml.X(0)), qml.expval(qml.X(1))

        check_methods(qf, {0: "F", 1: "F"})

        def qf(x, y):
            qml.Kerr(y, wires=[1])
            qml.Displacement(x, 0, wires=[0])
            qml.Beamsplitter(0.2, 1.7, wires=[0, 1])
            return qml.expval(qml.X(0)), qml.expval(qml.X(1))

        check_methods(qf, {0: "A", 1: "F"})

    def test_qnode_gradient_multiple_gate_parameters(self, tol):
        """Tests that gates with multiple free parameters yield correct gradients."""
        par = [0.5, 0.3, -0.7]

        def qf(x, y, z):
            qml.RX(0.4, wires=[0])
            qml.Rot(x, y, z, wires=[0])
            qml.RY(-0.2, wires=[0])
            return qml.expval(qml.PauliZ(0))

        dev = qml.device("default.qubit", wires=1)
        q = qml.QNode(qf, dev)
        value = q(*par)
        grad_A = q.jacobian(par, method="A")
        grad_F = q.jacobian(par, method="F")

        # analytic method works for every parameter
        assert q.grad_method_for_par == {0: "A", 1: "A", 2: "A"}
        # gradient has the correct shape and every element is nonzero
        assert grad_A.shape == (1, 3)
        assert np.count_nonzero(grad_A) == 3
        # the different methods agree
        assert np.allclose(grad_A, grad_F, atol=tol, rtol=0)

    def test_qnode_gradient_gate_with_two_parameters(self, tol):
        """Test that a gate with two parameters yields
        correct gradients"""
        def qf(r0, phi0, r1, phi1):
            qml.Squeezing(r0, phi0, wires=[0])
            qml.Squeezing(r1, phi1, wires=[0])
            return qml.expval(qml.NumberOperator(0))

        dev = qml.device('default.gaussian', wires=2)
        q = qml.QNode(qf, dev)

        par = [0.543, 0.123, 0.654, -0.629]

        grad_A = q.jacobian(par, method='A')
        grad_F = q.jacobian(par, method='F')

        # the different methods agree
        assert np.allclose(grad_A, grad_F, atol=tol, rtol=0)

    def test_qnode_gradient_repeated_gate_parameters(self, tol):
        """Tests that repeated use of a free parameter in a
        multi-parameter gate yield correct gradients."""
        par = [0.8, 1.3]

        def qf(x, y):
            qml.RX(np.pi / 4, wires=[0])
            qml.Rot(y, x, 2 * x, wires=[0])
            return qml.expval(qml.PauliX(0))

        dev = qml.device("default.qubit", wires=1)
        q = qml.QNode(qf, dev)
        grad_A = q.jacobian(par, method="A")
        grad_F = q.jacobian(par, method="F")

        # the different methods agree
        assert np.allclose(grad_A, grad_F, atol=tol, rtol=0)

    def test_qnode_gradient_parameters_inside_array(self, tol):
        """Tests that free parameters inside an array passed to
        an Operation yield correct gradients."""
        par = [0.8, 1.3]

        def qf(x, y):
            qml.RX(x, wires=[0])
            qml.RY(x, wires=[0])
            return qml.expval(qml.Hermitian(np.diag([y, 1]), 0))

        dev = qml.device("default.qubit", wires=1)
        q = qml.QNode(qf, dev)
        grad = q.jacobian(par)
        grad_F = q.jacobian(par, method="F")

        # par[0] can use the 'A' method, par[1] cannot
        assert q.grad_method_for_par == {0: "A", 1: "F"}
        # the different methods agree
        assert np.allclose(grad, grad_F, atol=tol, rtol=0)

    def test_array_parameters_autograd(self, tol):
        """Test that gradients of array parameters give
        same results as positional arguments."""

        a, b, c = 0.5, 0.54, 0.3

        def ansatz(x, y, z):
            qml.QubitStateVector(np.array([1, 0, 1, 1]) / np.sqrt(3), wires=[0, 1])
            qml.Rot(x, y, z, wires=0)
            qml.CNOT(wires=[0, 1])
            return qml.expval(qml.PauliZ(0))

        def circuit1(x, y, z):
            return ansatz(x, y, z)

        def circuit2(x, array):
            return ansatz(x, array[0], array[1])

        def circuit3(array):
            return ansatz(*array)

        dev = qml.device("default.qubit", wires=2)
        circuit1 = qml.QNode(circuit1, dev)
        grad1 = qml.grad(circuit1, argnum=[0, 1, 2])

        positional_grad = circuit1.jacobian([a, b, c])
        positional_autograd = grad1(a, b, c)
        assert np.allclose(positional_grad, positional_autograd, atol=tol, rtol=0)

        circuit2 = qml.QNode(circuit2, dev)
        grad2 = qml.grad(circuit2, argnum=[0, 1])

        circuit3 = qml.QNode(circuit3, dev)
        grad3 = qml.grad(circuit3, argnum=0)

        array_grad = circuit3.jacobian([np.array([a, b, c])])
        array_autograd = grad3(np.array([a, b, c]))
        assert np.allclose(array_grad, array_autograd, atol=tol, rtol=0)

    @staticmethod
    def expected_jacobian(x, y, z):
        dw0dx = 2 / 3 * np.sin(x) * np.sin(y)
        dw0dy = 1 / 3 * (np.sin(y) - 2 * np.cos(x) * np.cos(y))
        dw0dz = 0

        dw1dx = -2 / 3 * np.cos(x) * np.sin(y)
        dw1dy = -2 / 3 * np.cos(y) * np.sin(x)
        dw1dz = 0

        return np.array([[dw0dx, dw0dy, dw0dz], [dw1dx, dw1dy, dw1dz]])

    def test_multiple_expectation_jacobian_positional(self, tol):
        """Tests that qnodes using positional arguments return
        correct gradients for multiple expectation values."""
        a, b, c = 0.5, 0.54, 0.3

        def circuit(x, y, z):
            qml.QubitStateVector(np.array([1, 0, 1, 1]) / np.sqrt(3), wires=[0, 1])
            qml.Rot(x, y, z, wires=0)
            qml.CNOT(wires=[0, 1])
            return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliY(1))

        dev = qml.device("default.qubit", wires=2)
        circuit = qml.QNode(circuit, dev)

        # compare our manual Jacobian computation to theoretical result
        # Note: circuit.jacobian actually returns a full jacobian in this case
        res = circuit.jacobian(np.array([a, b, c]))
        assert np.allclose(self.expected_jacobian(a, b, c), res, atol=tol, rtol=0)

        # compare our manual Jacobian computation to autograd
        # not sure if this is the intended usage of jacobian
        jac0 = qml.jacobian(circuit, 0)
        jac1 = qml.jacobian(circuit, 1)
        jac2 = qml.jacobian(circuit, 2)
        res = np.stack([jac0(a, b, c), jac1(a, b, c), jac2(a, b, c)]).T

        assert np.allclose(self.expected_jacobian(a, b, c), res, atol=tol, rtol=0)

        #compare with what we get if argnum is a list
        res2 = qml.jacobian(circuit, argnum=[0, 1, 2])(a, b, c)
        assert np.allclose(res, res2, atol=tol, rtol=0)

    def test_multiple_expectation_jacobian_array(self, tol):
        """Tests that qnodes using an array argument return correct gradients
        for multiple expectation values."""
        a, b, c = 0.5, 0.54, 0.3

        def circuit(weights):
            qml.QubitStateVector(np.array([1, 0, 1, 1]) / np.sqrt(3), wires=[0, 1])
            qml.Rot(weights[0], weights[1], weights[2], wires=0)
            qml.CNOT(wires=[0, 1])
            return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliY(1))

        dev = qml.device("default.qubit", wires=2)
        circuit = qml.QNode(circuit, dev)

        res = circuit.jacobian([np.array([a, b, c])])
        assert np.allclose(self.expected_jacobian(a, b, c), res, atol=tol, rtol=0)

        jac = qml.jacobian(circuit, 0)
        res = jac(np.array([a, b, c]))
        assert np.allclose(self.expected_jacobian(a, b, c), res, atol=tol, rtol=0)

    def test_keywordarg_not_differentiated(self, tol):
        """Tests that qnodes do not differentiate w.r.t. keyword arguments."""
        a, b = 0.5, 0.54

        def circuit1(weights, x=0.3):
            qml.QubitStateVector(np.array([1, 0, 1, 1]) / np.sqrt(3), wires=[0, 1])
            qml.Rot(weights[0], weights[1], x, wires=0)
            qml.CNOT(wires=[0, 1])
            return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliY(1))

        dev = qml.device("default.qubit", wires=2)
        circuit1 = qml.QNode(circuit1, dev)

        def circuit2(weights):
            qml.QubitStateVector(np.array([1, 0, 1, 1]) / np.sqrt(3), wires=[0, 1])
            qml.Rot(weights[0], weights[1], 0.3, wires=0)
            qml.CNOT(wires=[0, 1])
            return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliY(1))

        circuit2 = qml.QNode(circuit2, dev)

        res1 = circuit1.jacobian([np.array([a, b])])
        res2 = circuit2.jacobian([np.array([a, b])])

        assert np.allclose(res1, res2, atol=tol, rtol=0)

    def test_differentiate_all_positional(self, tol):
        """Tests that all positional arguments are differentiated."""

        def circuit1(a, b, c):
            qml.RX(a, wires=0)
            qml.RX(b, wires=1)
            qml.RX(c, wires=2)
            return tuple(qml.expval(qml.PauliZ(idx)) for idx in range(3))

        dev = qml.device("default.qubit", wires=3)
        circuit1 = qml.QNode(circuit1, dev)

        vals = np.array([np.pi, np.pi / 2, np.pi / 3])
        circuit_output = circuit1(*vals)
        expected_output = np.cos(vals)
        assert np.allclose(circuit_output, expected_output, atol=tol, rtol=0)

        # circuit jacobians
        circuit_jacobian = circuit1.jacobian(vals)
        expected_jacobian = -np.diag(np.sin(vals))
        assert np.allclose(circuit_jacobian, expected_jacobian, atol=tol, rtol=0)

    def test_differentiate_first_positional(self, tol):
        """Tests that the first positional arguments are differentiated."""

        def circuit2(a, b):
            qml.RX(a, wires=0)
            return qml.expval(qml.PauliZ(0))

        dev = qml.device("default.qubit", wires=2)
        circuit2 = qml.QNode(circuit2, dev)

        a = 0.7418
        b = -5.0
        circuit_output = circuit2(a, b)
        expected_output = np.cos(a)
        assert np.allclose(circuit_output, expected_output, atol=tol, rtol=0)

        # circuit jacobians
        circuit_jacobian = circuit2.jacobian([a, b])
        expected_jacobian = np.array([[-np.sin(a), 0]])
        assert np.allclose(circuit_jacobian, expected_jacobian, atol=tol, rtol=0)

    def test_differentiate_second_positional(self, tol):
        """Tests that the second positional arguments are differentiated."""

        def circuit3(a, b):
            qml.RX(b, wires=0)
            return qml.expval(qml.PauliZ(0))

        dev = qml.device("default.qubit", wires=2)
        circuit3 = qml.QNode(circuit3, dev)

        a = 0.7418
        b = -5.0
        circuit_output = circuit3(a, b)
        expected_output = np.cos(b)
        assert np.allclose(circuit_output, expected_output, atol=tol, rtol=0)

        # circuit jacobians
        circuit_jacobian = circuit3.jacobian([a, b])
        expected_jacobian = np.array([[0, -np.sin(b)]])
        assert np.allclose(circuit_jacobian, expected_jacobian, atol=tol, rtol=0)

    def test_differentiate_second_third_positional(self, tol):
        """Tests that the second and third positional arguments are differentiated."""

        def circuit4(a, b, c):
            qml.RX(b, wires=0)
            qml.RX(c, wires=1)
            return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliZ(1))

        dev = qml.device("default.qubit", wires=2)
        circuit4 = qml.QNode(circuit4, dev)

        a = 0.7418
        b = -5.0
        c = np.pi / 7
        circuit_output = circuit4(a, b, c)
        expected_output = np.array([[np.cos(b), np.cos(c)]])
        assert np.allclose(circuit_output, expected_output, atol=tol, rtol=0)

        # circuit jacobians
        circuit_jacobian = circuit4.jacobian([a, b, c])
        expected_jacobian = np.array([[0.0, -np.sin(b), 0.0], [0.0, 0.0, -np.sin(c)]])
        assert np.allclose(circuit_jacobian, expected_jacobian, atol=tol, rtol=0)

    def test_differentiate_positional_multidim(self, tol):
        """Tests that all positional arguments are differentiated
        when they are multidimensional."""

        def circuit(a, b):
            qml.RX(a[0], wires=0)
            qml.RX(a[1], wires=1)
            qml.RX(b[2, 1], wires=2)
            return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliZ(1)), qml.expval(qml.PauliZ(2))

        dev = qml.device("default.qubit", wires=3)
        circuit = qml.QNode(circuit, dev)

        a = np.array([-np.sqrt(2), -0.54])
        b = np.array([np.pi / 7] * 6).reshape([3, 2])
        circuit_output = circuit(a, b)
        expected_output = np.cos(np.array([[a[0], a[1], b[-1, 0]]]))
        assert np.allclose(circuit_output, expected_output, atol=tol, rtol=0)

        # circuit jacobians
        circuit_jacobian = circuit.jacobian([a, b])
        expected_jacobian = np.array(
            [
                [-np.sin(a[0])] + [0.0] * 7,  # expval 0
                [0.0, -np.sin(a[1])] + [0.0] * 6,  # expval 1
                [0.0] * 2 + [0.0] * 5 + [-np.sin(b[2, 1])],
            ]
        )  # expval 2
        assert np.allclose(circuit_jacobian, expected_jacobian, atol=tol, rtol=0)

    def test_controlled_RX_gradient(self, tol):
        """Test gradient of controlled RX gate"""
        dev = qml.device("default.qubit", wires=2)

        @qml.qnode(dev)
        def circuit(x):
            qml.PauliX(wires=0)
            qml.CRX(x, wires=[0, 1])
            return qml.expval(qml.PauliZ(0))

        a = 0.542  # any value of a should give zero gradient

        # get the analytic gradient
        gradA = circuit.jacobian([a], method="A")
        # get the finite difference gradient
        gradF = circuit.jacobian([a], method="F")

        # the expected gradient
        expected = 0

        assert np.allclose(gradF, expected, atol=tol, rtol=0)
        assert np.allclose(gradA, expected, atol=tol, rtol=0)

        @qml.qnode(dev)
        def circuit1(x):
            qml.RX(x, wires=0)
            qml.CRX(x, wires=[0, 1])
            return qml.expval(qml.PauliZ(0))

        b = 0.123  # gradient is -sin(x)

        # get the analytic gradient
        gradA = circuit1.jacobian([b], method="A")
        # get the finite difference gradient
        gradF = circuit1.jacobian([b], method="F")

        # the expected gradient
        expected = -np.sin(b)

        assert np.allclose(gradF, expected, atol=tol, rtol=0)
        assert np.allclose(gradA, expected, atol=tol, rtol=0)

    def test_controlled_RY_gradient(self, tol):
        """Test gradient of controlled RY gate"""
        dev = qml.device("default.qubit", wires=2)

        @qml.qnode(dev)
        def circuit(x):
            qml.PauliX(wires=0)
            qml.CRY(x, wires=[0, 1])
            return qml.expval(qml.PauliZ(0))

        a = 0.542  # any value of a should give zero gradient

        # get the analytic gradient
        gradA = circuit.jacobian([a], method="A")
        # get the finite difference gradient
        gradF = circuit.jacobian([a], method="F")

        # the expected gradient
        expected = 0

        assert np.allclose(gradF, expected, atol=tol, rtol=0)
        assert np.allclose(gradA, expected, atol=tol, rtol=0)

        @qml.qnode(dev)
        def circuit1(x):
            qml.RX(x, wires=0)
            qml.CRY(x, wires=[0, 1])
            return qml.expval(qml.PauliZ(0))

        b = 0.123  # gradient is -sin(x)

        # get the analytic gradient
        gradA = circuit1.jacobian([b], method="A")
        # get the finite difference gradient
        gradF = circuit1.jacobian([b], method="F")

        # the expected gradient
        expected = -np.sin(b)

        assert np.allclose(gradF, expected, atol=tol, rtol=0)
        assert np.allclose(gradA, expected, atol=tol, rtol=0)

    def test_controlled_RZ_gradient(self, tol):
        """Test gradient of controlled RZ gate"""
        dev = qml.device("default.qubit", wires=2)

        @qml.qnode(dev)
        def circuit(x):
            qml.PauliX(wires=0)
            qml.CRZ(x, wires=[0, 1])
            return qml.expval(qml.PauliZ(0))

        a = 0.542  # any value of a should give zero gradient

        # get the analytic gradient
        gradA = circuit.jacobian([a], method="A")
        # get the finite difference gradient
        gradF = circuit.jacobian([a], method="F")

        # the expected gradient
        expected = 0

        assert np.allclose(gradF, expected, atol=tol, rtol=0)
        assert np.allclose(gradA, expected, atol=tol, rtol=0)

        @qml.qnode(dev)
        def circuit1(x):
            qml.RX(x, wires=0)
            qml.CRZ(x, wires=[0, 1])
            return qml.expval(qml.PauliZ(0))

        b = 0.123  # gradient is -sin(x)

        # get the analytic gradient
        gradA = circuit1.jacobian([b], method="A")
        # get the finite difference gradient
        gradF = circuit1.jacobian([b], method="F")

        # the expected gradient
        expected = -np.sin(b)

        assert np.allclose(gradF, expected, atol=tol, rtol=0)
        assert np.allclose(gradA, expected, atol=tol, rtol=0)


gradient_test_data = [
    (0.5, -0.1),
    (0.0, np.pi),
    (-3.6, -3.6),
    (1.0, 2.5),
]


dev = qml.device("default.qubit", wires=2)


@qml.qnode(dev)
def f(x):
    qml.RX(x, wires=0)
    return qml.expval(qml.PauliZ(0))


@qml.qnode(dev)
def g(y):
    qml.RY(y, wires=0)
    return qml.expval(qml.PauliX(0))


class TestMultiQNodeGradients:
    """Multi Qnode gradient tests."""

    @pytest.mark.parametrize("x, y", gradient_test_data)
    def test_add_qnodes_gradient(self, x, y):
        """Test the gradient of addition of two QNode circuits"""

        def add(a, b):
            return a + b

        a = f(x)
        b = g(y)

        # addition
        assert qml.grad(add, argnum=0)(a, b) == 1.0
        assert qml.grad(add, argnum=1)(a, b) == 1.0

        # same value added to itself; autograd doesn't distinguish inputs
        assert qml.grad(add, argnum=0)(a, a) == 1.0

    @pytest.mark.parametrize("x, y", gradient_test_data)
    def test_subtract_qnodes_gradient(self, x, y):
        """Test the gradient of subtraction of two QNode circuits"""

        def subtract(a, b):
            return a - b

        a = f(x)
        b = g(y)

        # subtraction
        assert qml.grad(subtract, argnum=0)(a, b) == 1.0
        assert qml.grad(subtract, argnum=1)(a, b) == -1.0

    @pytest.mark.parametrize("x, y", gradient_test_data)
    def test_multiply_qnodes_gradient(self, x, y):
        """Test the gradient of multiplication of two QNode circuits"""

        def mult(a, b):
            return a * b

        a = f(x)
        b = g(y)

        # multipication
        assert qml.grad(mult, argnum=0)(a, b) == b
        assert qml.grad(mult, argnum=1)(a, b) == a

    @pytest.mark.parametrize("x, y", gradient_test_data)
    def test_division_qnodes_gradient(self, x, y):
        """Test the gradient of division of two QNode circuits"""

        def div(a, b):
            return a / b

        a = f(x)
        b = g(y)

        # division
        assert qml.grad(div, argnum=0)(a, b) == 1 / b
        assert qml.grad(div, argnum=1)(a, b) == -a / b ** 2

    @pytest.mark.parametrize("x, y", gradient_test_data)
    def test_composing_qnodes_gradient(self, x, y):
        """Test the gradient of composing of two QNode circuits"""

        def compose(f, x):
            return f(x)

        a = f(x)
        b = g(y)

        # composition
        assert qml.grad(compose, argnum=1)(f, x) == qml.grad(f, argnum=0)(x)
        assert qml.grad(compose, argnum=1)(f, a) == qml.grad(f, argnum=0)(a)
        assert qml.grad(compose, argnum=1)(f, b) == qml.grad(f, argnum=0)(b)

class TestQNodeVariance:
    """Qnode variance tests."""

    def test_involutory_variance(self, tol):
        """Tests qubit observable that are involutory"""
        dev = qml.device("default.qubit", wires=1)

        @qml.qnode(dev)
        def circuit(a):
            qml.RX(a, wires=0)
            return qml.var(qml.PauliZ(0))

        a = 0.54
        var = circuit(a)
        expected = 1 - np.cos(a) ** 2
        assert np.allclose(var, expected, atol=tol, rtol=0)

        # circuit jacobians
        gradA = circuit.jacobian([a], method="A")
        gradF = circuit.jacobian([a], method="F")
        expected = 2 * np.sin(a) * np.cos(a)
        assert np.allclose(gradF, expected, atol=tol, rtol=0)
        assert np.allclose(gradA, expected, atol=tol, rtol=0)

    def test_non_involutory_variance(self, tol):
        """Tests a qubit Hermitian observable that is not involutory"""
        dev = qml.device("default.qubit", wires=1)

        A = np.array([[4, -1 + 6j], [-1 - 6j, 2]])

        @qml.qnode(dev)
        def circuit(a):
            qml.RX(a, wires=0)
            return qml.var(qml.Hermitian(A, 0))

        a = 0.54
        var = circuit(a)
        expected = (39 / 2) - 6 * np.sin(2 * a) + (35 / 2) * np.cos(2 * a)
        assert np.allclose(var, expected, atol=tol, rtol=0)

        # circuit jacobians
        gradA = circuit.jacobian([a], method="A")
        gradF = circuit.jacobian([a], method="F")
        expected = -35 * np.sin(2 * a) - 12 * np.cos(2 * a)
        assert np.allclose(gradA, expected, atol=tol, rtol=0)
        assert np.allclose(gradF, expected, atol=tol, rtol=0)

    def test_fanout(self, tol):
        """Tests qubit observable with repeated parameters"""
        dev = qml.device("default.qubit", wires=1)

        @qml.qnode(dev)
        def circuit(a):
            qml.RX(a, wires=0)
            qml.RY(a, wires=0)
            return qml.var(qml.PauliZ(0))

        a = 0.54
        var = circuit(a)
        expected = 0.5 * np.sin(a) ** 2 * (np.cos(2 * a) + 3)
        assert np.allclose(var, expected, atol=tol, rtol=0)

        # circuit jacobians
        gradA = circuit.jacobian([a], method="A")
        gradF = circuit.jacobian([a], method="F")
        expected = 4 * np.sin(a) * np.cos(a) ** 3
        assert np.allclose(gradA, expected, atol=tol, rtol=0)
        assert np.allclose(gradF, expected, atol=tol, rtol=0)

    def test_expval_and_variance(self, tol):
        """Test that the qnode works for a combination of expectation
        values and variances"""
        dev = qml.device("default.qubit", wires=3)

        @qml.qnode(dev)
        def circuit(a, b, c):
            qml.RX(a, wires=0)
            qml.RY(b, wires=1)
            qml.CNOT(wires=[1, 2])
            qml.RX(c, wires=2)
            qml.CNOT(wires=[0, 1])
            qml.RZ(c, wires=2)
            return qml.var(qml.PauliZ(0)), qml.expval(qml.PauliZ(1)), qml.var(qml.PauliZ(2))

        a = 0.54
        b = -0.423
        c = 0.123

        var = circuit(a, b, c)
        expected = np.array(
            [
                np.sin(a) ** 2,
                np.cos(a) * np.cos(b),
                0.25 * (3 - 2 * np.cos(b) ** 2 * np.cos(2 * c) - np.cos(2 * b)),
            ]
        )
        assert np.allclose(var, expected, atol=tol, rtol=0)

        # # circuit jacobians
        gradA = circuit.jacobian([a, b, c], method="A")
        gradF = circuit.jacobian([a, b, c], method="F")
        expected = np.array(
            [
                [2 * np.cos(a) * np.sin(a), -np.cos(b) * np.sin(a), 0],
                [
                    0,
                    -np.cos(a) * np.sin(b),
                    0.5 * (2 * np.cos(b) * np.cos(2 * c) * np.sin(b) + np.sin(2 * b)),
                ],
                [0, 0, np.cos(b) ** 2 * np.sin(2 * c)],
            ]
        ).T
        assert np.allclose(gradF, expected, atol=tol, rtol=0)
        assert np.allclose(gradA, expected, atol=tol, rtol=0)

    def test_first_order_cv(self, tol):
        """Test variance of a first order CV expectation value"""
        dev = qml.device("default.gaussian", wires=1)

        @qml.qnode(dev)
        def circuit(r, phi):
            qml.Squeezing(r, 0, wires=0)
            qml.Rotation(phi, wires=0)
            return qml.var(qml.X(0))

        r = 0.543
        phi = -0.654

        var = circuit(r, phi)
        expected = np.exp(2 * r) * np.sin(phi) ** 2 + np.exp(-2 * r) * np.cos(phi) ** 2
        assert np.allclose(var, expected, atol=tol, rtol=0)

        # circuit jacobians
        gradA = circuit.jacobian([r, phi], method="A")
        gradF = circuit.jacobian([r, phi], method="F")
        expected = np.array(
            [
                2 * np.exp(2 * r) * np.sin(phi) ** 2 - 2 * np.exp(-2 * r) * np.cos(phi) ** 2,
                2 * np.sinh(2 * r) * np.sin(2 * phi),
            ]
        )
        assert np.allclose(gradA, expected, atol=tol, rtol=0)
        assert np.allclose(gradF, expected, atol=tol, rtol=0)

    def test_second_order_cv(self, tol):
        """Test variance of a second order CV expectation value"""
        dev = qml.device("default.gaussian", wires=1)

        @qml.qnode(dev)
        def circuit(n, a):
            qml.ThermalState(n, wires=0)
            qml.Displacement(a, 0, wires=0)
            return qml.var(qml.NumberOperator(0))

        n = 0.12
        a = 0.765

        var = circuit(n, a)
        expected = n ** 2 + n + np.abs(a) ** 2 * (1 + 2 * n)
        assert np.allclose(var, expected, atol=tol, rtol=0)

        # circuit jacobians
        gradF = circuit.jacobian([n, a], method="F")
        expected = np.array([2 * a ** 2 + 2 * n + 1, 2 * a * (2 * n + 1)])
        assert np.allclose(gradF, expected, atol=tol, rtol=0)

    def test_error_analytic_second_order_cv(self):
        """Test exception raised if attempting to use a second
        order observable to compute the variance derivative analytically"""
        dev = qml.device("default.gaussian", wires=1)

        @qml.qnode(dev)
        def circuit(a):
            qml.Displacement(a, 0, wires=0)
            return qml.var(qml.NumberOperator(0))

        with pytest.raises(ValueError, match=r"cannot be used with the parameter\(s\) \{0\}"):
            circuit.jacobian([1.0], method="A")


class TestMetricTensor:
    """Tests for metric tensor subcircuit construction and evaluation"""

    def test_no_generator(self):
        """Test exception is raised if subcircuit contains an
        operation with no generator"""
        dev = qml.device('default.qubit', wires=1)

        def circuit(a):
            qml.Rot(a, 0, 0, wires=0)
            return qml.expval(qml.PauliX(0))

        circuit = qml.QNode(circuit, dev)

        with pytest.raises(QuantumFunctionError, match="has no defined generator"):
            circuit.construct_metric_tensor([1])

    def test_generator_no_expval(self, monkeypatch):
        """Test exception is raised if subcircuit contains an
        operation with generator object that is not an observable"""
        dev = qml.device('default.qubit', wires=1)

        def circuit(a):
            qml.RX(a, wires=0)
            return qml.expval(qml.PauliX(0))

        circuit = qml.QNode(circuit, dev)

        with monkeypatch.context() as m:
            m.setattr('pennylane.RX.generator', [qml.RX, 1])

            with pytest.raises(QuantumFunctionError, match="no corresponding observable"):
                circuit.construct_metric_tensor([1])

    def test_construct_subcircuit(self):
        """Test correct subcircuits constructed"""
        dev = qml.device('default.qubit', wires=2)

        def circuit(a, b, c):
            qml.RX(a, wires=0)
            qml.RY(b, wires=0)
            qml.CNOT(wires=[0, 1])
            qml.PhaseShift(c, wires=1)
            return qml.expval(qml.PauliX(0))

        circuit = qml.QNode(circuit, dev)

        circuit.construct_metric_tensor([1, 1, 1])
        res = circuit._metric_tensor_subcircuits

        # first parameter subcircuit
        assert len(res[(0,)]['queue']) == 0
        assert res[(0,)]['scale'] == [-0.5]
        assert isinstance(res[(0,)]['observable'][0], qml.PauliX)

        # second parameter subcircuit
        assert len(res[(1,)]['queue']) == 1
        assert res[(1,)]['scale'] == [-0.5]
        assert isinstance(res[(1,)]['queue'][0], qml.RX)
        assert isinstance(res[(1,)]['observable'][0], qml.PauliY)

        # third parameter subcircuit
        assert len(res[(2,)]['queue']) == 3
        assert res[(2,)]['scale'] == [1]
        assert isinstance(res[(2,)]['queue'][0], qml.RX)
        assert isinstance(res[(2,)]['queue'][1], qml.RY)
        assert isinstance(res[(2,)]['queue'][2], qml.CNOT)
        assert isinstance(res[(2,)]['observable'][0], qml.Hermitian)
        assert np.all(res[(2,)]['observable'][0].params[0] == qml.PhaseShift.generator[0])

    def test_construct_subcircuit_layers(self):
        """Test correct subcircuits constructed
        when a layer structure exists"""
        dev = qml.device('default.qubit', wires=3)

        def circuit(params):
            # section 1
            qml.RX(params[0], wires=0)
            # section 2
            qml.RY(params[1], wires=0)
            qml.CNOT(wires=[0, 1])
            qml.CNOT(wires=[1, 2])
            # section 3
            qml.RX(params[2], wires=0)
            qml.RY(params[3], wires=1)
            qml.RZ(params[4], wires=2)
            qml.CNOT(wires=[0, 1])
            qml.CNOT(wires=[1, 2])
            # section 4
            qml.RX(params[5], wires=0)
            qml.RY(params[6], wires=1)
            qml.RZ(params[7], wires=2)
            qml.CNOT(wires=[0, 1])
            qml.CNOT(wires=[1, 2])
            return qml.expval(qml.PauliX(0))

        circuit = qml.QNode(circuit, dev)

        params = np.ones([8])
        circuit.construct_metric_tensor([params])
        res = circuit._metric_tensor_subcircuits

        # this circuit should split into 4 independent
        # sections or layers when constructing subcircuits
        assert len(res) == 4

        # first layer subcircuit
        layer = res[(0,)]
        assert len(layer['queue']) == 0
        assert len(layer['observable']) == 1
        assert isinstance(layer['observable'][0], qml.PauliX)

        # second layer subcircuit
        layer = res[(1,)]
        assert len(layer['queue']) == 1
        assert len(layer['observable']) == 1
        assert isinstance(layer['queue'][0], qml.RX)
        assert isinstance(layer['observable'][0], qml.PauliY)

        # third layer subcircuit
        layer = res[(2, 3, 4)]
        assert len(layer['queue']) == 4
        assert len(layer['observable']) == 3
        assert isinstance(layer['queue'][0], qml.RX)
        assert isinstance(layer['queue'][1], qml.RY)
        assert isinstance(layer['queue'][2], qml.CNOT)
        assert isinstance(layer['queue'][3], qml.CNOT)
        assert isinstance(layer['observable'][0], qml.PauliX)
        assert isinstance(layer['observable'][1], qml.PauliY)
        assert isinstance(layer['observable'][2], qml.PauliZ)

        # fourth layer subcircuit
        layer = res[(5, 6, 7)]
        assert len(layer['queue']) == 9
        assert len(layer['observable']) == 3
        assert isinstance(layer['queue'][0], qml.RX)
        assert isinstance(layer['queue'][1], qml.RY)
        assert isinstance(layer['queue'][2], qml.CNOT)
        assert isinstance(layer['queue'][3], qml.CNOT)
        assert isinstance(layer['queue'][4], qml.RX)
        assert isinstance(layer['queue'][5], qml.RY)
        assert isinstance(layer['queue'][6], qml.RZ)
        assert isinstance(layer['queue'][7], qml.CNOT)
        assert isinstance(layer['queue'][8], qml.CNOT)
        assert isinstance(layer['observable'][0], qml.PauliX)
        assert isinstance(layer['observable'][1], qml.PauliY)
        assert isinstance(layer['observable'][2], qml.PauliZ)

    def test_evaluate_subcircuits(self, tol):
        """Test subcircuits evaluate correctly"""
        dev = qml.device('default.qubit', wires=2)

        def circuit(a, b, c):
            qml.RX(a, wires=0)
            qml.RY(b, wires=0)
            qml.CNOT(wires=[0, 1])
            qml.PhaseShift(c, wires=1)
            return qml.expval(qml.PauliX(0))

        # construct subcircuits
        circuit = qml.QNode(circuit, dev)
        circuit.construct_metric_tensor([1, 1, 1])

        a = 0.432
        b = 0.12
        c = -0.432

        # evaluate subcircuits
        circuit.metric_tensor(a, b, c)

        # first parameter subcircuit
        res = circuit._metric_tensor_subcircuits[(0,)]['result']
        expected = 0.25
        assert np.allclose(res, expected, atol=tol, rtol=0)

        # second parameter subcircuit
        res = circuit._metric_tensor_subcircuits[(1,)]['result']
        expected = np.cos(a)**2/4
        assert np.allclose(res, expected, atol=tol, rtol=0)

        # third parameter subcircuit
        res = circuit._metric_tensor_subcircuits[(2,)]['result']
        expected = (3-2*np.cos(a)**2*np.cos(2*b)-np.cos(2*a))/16
        assert np.allclose(res, expected, atol=tol, rtol=0)

    def test_evaluate_diag_metric_tensor(self, tol):
        """Test that a diagonal metric tensor evaluates correctly"""
        dev = qml.device('default.qubit', wires=2)

        @qml.qnode(dev)
        def circuit(a, b, c):
            qml.RX(a, wires=0)
            qml.RY(b, wires=0)
            qml.CNOT(wires=[0, 1])
            qml.PhaseShift(c, wires=1)
            return qml.expval(qml.PauliX(0))

        a = 0.432
        b = 0.12
        c = -0.432

        # evaluate metric tensor
        g = circuit.metric_tensor(a, b, c)

        # check that the metric tensor is correct
        expected = np.array([1, np.cos(a)**2, (3-2*np.cos(a)**2*np.cos(2*b)-np.cos(2*a))/4])/4
        assert np.allclose(g, np.diag(expected), atol=tol, rtol=0)

    @pytest.fixture
    def sample_circuit(self):
        """Sample variational circuit fixture used in the
        next couple of tests"""
        dev = qml.device('default.qubit', wires=3)

        def non_parametrized_layer(a, b, c):
            qml.RX(a, wires=0)
            qml.RX(b, wires=1)
            qml.RX(c, wires=1)
            qml.CNOT(wires=[0, 1])
            qml.CNOT(wires=[1, 2])
            qml.RZ(a, wires=0)
            qml.Hadamard(wires=1)
            qml.CNOT(wires=[0, 1])
            qml.RZ(b, wires=1)
            qml.Hadamard(wires=0)

        a = 0.5
        b = 0.1
        c = 0.5

        @qml.qnode(dev)
        def final(x, y, z, h, g, f):
            non_parametrized_layer(a, b, c)
            qml.RX(x, wires=0)
            qml.RY(y, wires=1)
            qml.RZ(z, wires=2)
            non_parametrized_layer(a, b, c)
            qml.RY(f, wires=1)
            qml.RZ(g, wires=2)
            qml.RX(h, wires=1)
            return qml.expval(qml.PauliX(0))

        return dev, final, non_parametrized_layer, a, b, c

    def test_evaluate_block_diag_metric_tensor(self, sample_circuit, tol):
        """Test that a block diagonal metric tensor evaluates correctly,
        by comparing it to a known analytic result as well as numerical
        computation."""
        dev, circuit, non_parametrized_layer, a, b, c = sample_circuit

        params = [-0.282203, 0.145554, 0.331624, -0.163907, 0.57662, 0.081272]
        x, y, z, h, g, f = params

        G = circuit.metric_tensor(x, y, z, h, g, f)

        # ============================================
        # Test block diag metric tensor of first layer is correct.
        # We do this by comparing against the known analytic result.
        # First layer includes the non_parametrized_layer,
        # followed by observables corresponding to generators of:
        #   qml.RX(x, wires=0)
        #   qml.RY(y, wires=1)
        #   qml.RZ(z, wires=2)

        G1 = np.zeros([3, 3])

        # diag elements
        G1[0, 0] = np.sin(a)**2/4
        G1[1, 1] = (
            16 * np.cos(a) ** 2 * np.sin(b) ** 3 * np.cos(b) * np.sin(2 * c)
            + np.cos(2 * b) * (2 - 8 * np.cos(a) ** 2 * np.sin(b) ** 2 * np.cos(2 * c))
            + np.cos(2 * (a - b))
            + np.cos(2 * (a + b))
            - 2 * np.cos(2 * a)
            + 14
        ) / 64
        G1[2, 2] = (3-np.cos(2*a)-2*np.cos(a)**2*np.cos(2*(b+c)))/16

        # off diag elements
        G1[0, 1] = np.sin(a)**2 * np.sin(b) * np.cos(b+c)/4
        G1[0, 2] = np.sin(a)**2 * np.cos(b+c)/4
        G1[1, 2] = -np.sin(b) * (
            np.cos(2 * (a - b - c))
            + np.cos(2 * (a + b + c))
            + 2 * np.cos(2 * a)
            + 2 * np.cos(2 * (b + c))
            - 6
        ) / 32

        G1[1, 0] = G1[0, 1]
        G1[2, 0] = G1[0, 2]
        G1[2, 1] = G1[1, 2]

        assert np.allclose(G[:3, :3], G1, atol=tol, rtol=0)

        # =============================================
        # Test block diag metric tensor of second layer is correct.
        # We do this by computing the required expectation values
        # numerically.
        # The second layer includes the non_parametrized_layer,
        # RX, RY, RZ gates (x, y, z params), a 2nd non_parametrized_layer,
        # followed by the qml.RY(f, wires=2) operation.
        #
        # Observable is simply generator of:
        #   qml.RY(f, wires=2)
        #
        # Note: since this layer only consists of a single parameter,
        # only need to compute a single diagonal element.

        @qml.qnode(dev)
        def layer2_diag(x, y, z, h, g, f):
            non_parametrized_layer(a, b, c)
            qml.RX(x, wires=0)
            qml.RY(y, wires=1)
            qml.RZ(z, wires=2)
            non_parametrized_layer(a, b, c)
            qml.RY(f, wires=2)
            return qml.var(qml.PauliX(1))

        G2 = layer2_diag(x, y, z, h, g, f)/4
        assert np.allclose(G[3:4, 3:4], G2, atol=tol, rtol=0)

        # =============================================
        # Test block diag metric tensor of third layer is correct.
        # We do this by computing the required expectation values
        # numerically using multiple circuits.
        # The second layer includes the non_parametrized_layer,
        # RX, RY, RZ gates (x, y, z params), and a 2nd non_parametrized_layer.
        #
        # Observables are the generators of:
        #   qml.RY(f, wires=1)
        #   qml.RZ(g, wires=2)
        G3 = np.zeros([2, 2])

        @qml.qnode(dev)
        def layer3_diag(x, y, z, h, g, f):
            non_parametrized_layer(a, b, c)
            qml.RX(x, wires=0)
            qml.RY(y, wires=1)
            qml.RZ(z, wires=2)
            non_parametrized_layer(a, b, c)
            return qml.var(qml.PauliZ(2)), qml.var(qml.PauliY(1))

        @qml.qnode(dev)
        def layer3_off_diag_first_order(x, y, z, h, g, f):
            non_parametrized_layer(a, b, c)
            qml.RX(x, wires=0)
            qml.RY(y, wires=1)
            qml.RZ(z, wires=2)
            non_parametrized_layer(a, b, c)
            return qml.expval(qml.PauliZ(2)), qml.expval(qml.PauliY(1))

        @qml.qnode(dev)
        def layer3_off_diag_second_order(x, y, z, h, g, f):
            non_parametrized_layer(a, b, c)
            qml.RX(x, wires=0)
            qml.RY(y, wires=1)
            qml.RZ(z, wires=2)
            non_parametrized_layer(a, b, c)
            return qml.expval(qml.Hermitian(np.kron(Z, Y), wires=[2, 1]))

        # calculate the diagonal terms
        varK0, varK1 = layer3_diag(x, y, z, h, g, f)
        G3[0, 0] = varK0/4
        G3[1, 1] = varK1/4

        # calculate the off-diagonal terms
        exK0, exK1 = layer3_off_diag_first_order(x, y, z, h, g, f)
        exK01 = layer3_off_diag_second_order(x, y, z, h, g, f)

        G3[0, 1] = (exK01 - exK0*exK1)/4
        G3[1, 0] = (exK01 - exK0*exK1)/4

        assert np.allclose(G[4:6, 4:6], G3, atol=tol, rtol=0)

        # ============================================
        # Finally, double check that the entire metric
        # tensor is as computed.

        G_expected = block_diag(G1, G2, G3)
        assert np.allclose(G, G_expected, atol=tol, rtol=0)

    def test_evaluate_diag_approx_metric_tensor(self, sample_circuit, tol):
        """Test that a metric tensor under the
        diagonal approximation evaluates correctly."""
        dev, circuit, non_parametrized_layer, a, b, c = sample_circuit
        params = [-0.282203, 0.145554, 0.331624, -0.163907, 0.57662, 0.081272]
        x, y, z, h, g, f = params

        G = circuit.metric_tensor(x, y, z, h, g, f, diag_approx=True)

        # ============================================
        # Test block diag metric tensor of first layer is correct.
        # We do this by comparing against the known analytic result.
        # First layer includes the non_parametrized_layer,
        # followed by observables corresponding to generators of:
        #   qml.RX(x, wires=0)
        #   qml.RY(y, wires=1)
        #   qml.RZ(z, wires=2)

        G1 = np.zeros([3, 3])

        # diag elements
        G1[0, 0] = np.sin(a)**2/4
        G1[1, 1] = (
            16 * np.cos(a) ** 2 * np.sin(b) ** 3 * np.cos(b) * np.sin(2 * c)
            + np.cos(2 * b) * (2 - 8 * np.cos(a) ** 2 * np.sin(b) ** 2 * np.cos(2 * c))
            + np.cos(2 * (a - b))
            + np.cos(2 * (a + b))
            - 2 * np.cos(2 * a)
            + 14
        ) / 64
        G1[2, 2] = (3-np.cos(2*a)-2*np.cos(a)**2*np.cos(2*(b+c)))/16

        assert np.allclose(G[:3, :3], G1, atol=tol, rtol=0)

        # =============================================
        # Test metric tensor of second layer is correct.
        # We do this by computing the required expectation values
        # numerically.
        # The second layer includes the non_parametrized_layer,
        # RX, RY, RZ gates (x, y, z params), a 2nd non_parametrized_layer,
        # followed by the qml.RY(f, wires=2) operation.
        #
        # Observable is simply generator of:
        #   qml.RY(f, wires=2)
        #
        # Note: since this layer only consists of a single parameter,
        # only need to compute a single diagonal element.

        @qml.qnode(dev)
        def layer2_diag(x, y, z, h, g, f):
            non_parametrized_layer(a, b, c)
            qml.RX(x, wires=0)
            qml.RY(y, wires=1)
            qml.RZ(z, wires=2)
            non_parametrized_layer(a, b, c)
            qml.RY(f, wires=2)
            return qml.var(qml.PauliX(1))

        G2 = layer2_diag(x, y, z, h, g, f)/4
        assert np.allclose(G[3:4, 3:4], G2, atol=tol, rtol=0)

        # =============================================
        # Test block diag metric tensor of third layer is correct.
        # We do this by computing the required expectation values
        # numerically using multiple circuits.
        # The second layer includes the non_parametrized_layer,
        # RX, RY, RZ gates (x, y, z params), and a 2nd non_parametrized_layer.
        #
        # Observables are the generators of:
        #   qml.RY(f, wires=1)
        #   qml.RZ(g, wires=2)
        G3 = np.zeros([2, 2])

        @qml.qnode(dev)
        def layer3_diag(x, y, z, h, g, f):
            non_parametrized_layer(a, b, c)
            qml.RX(x, wires=0)
            qml.RY(y, wires=1)
            qml.RZ(z, wires=2)
            non_parametrized_layer(a, b, c)
            return qml.var(qml.PauliZ(2)), qml.var(qml.PauliY(1))

        # calculate the diagonal terms
        varK0, varK1 = layer3_diag(x, y, z, h, g, f)
        G3[0, 0] = varK0/4
        G3[1, 1] = varK1/4

        assert np.allclose(G[4:6, 4:6], G3, atol=tol, rtol=0)

        # ============================================
        # Finally, double check that the entire metric
        # tensor is as computed.

        G_expected = block_diag(G1, G2, G3)
        assert np.allclose(G, G_expected, atol=tol, rtol=0)


class TestQNodeCacheing:
    """Tests for the QNode construction caching"""

    def test_no_caching(self):
        """Test that the circuit structure changes on
        subsequent evalutions with caching turned off
        """
        dev = qml.device("default.qubit", wires=2)

        def circuit(x, c=None):
            qml.RX(x, wires=0)

            for i in range(c):
                qml.RX(x, wires=i)

            return qml.expval(qml.PauliZ(0))

        circuit = qml.QNode(circuit, dev, cache=False)

        # first evaluation
        circuit(0, c=0)
        # check structure
        assert len(circuit.queue) == 1

        # second evaluation
        circuit(0, c=1)
        # check structure
        assert len(circuit.queue) == 2

    def test_caching(self):
        """Test that the circuit structure does not change on
        subsequent evalutions with caching turned on
        """
        dev = qml.device("default.qubit", wires=2)

        def circuit(x, c=None):
            qml.RX(x, wires=0)

            for i in range(c.val):
                qml.RX(x, wires=i)

            return qml.expval(qml.PauliZ(0))

        circuit = qml.QNode(circuit, dev, cache=True)

        # first evaluation
        circuit(0, c=0)
        # check structure
        assert len(circuit.queue) == 1

        # second evaluation
        circuit(0, c=1)
        # check structure
        assert len(circuit.queue) == 1



r"""
.. _barren_plateaus:

Barren plateaus in quantum neural networks
==========================================
*Author: Shahnawaz Ahmed (shahnawaz.ahmed95@gmail.com)*

In classical optimization, it is suggested that saddle
points, not local minima, provide a fundamental impediment
to rapid high-dimensional non-convex optimization
(Dauphin et al., 2014).

The problem of such barren plateaus manifests in a different
form in variational quantum circuits, which are at the heart
of techniques such as quantum neural networks or approximate
optimization e.g., QAOA (Quantum Adiabatic Optimization Algorithm)
which can be found in this `PennyLane QAOA tutorial
<https://pennylane.readthedocs.io/en/latest/tutorials/pennylane_run_qaoa_maxcut.html#qaoa-maxcut>`_.

While starting from a parameterized
random quantum circuit seems like a good unbiased choice if
we do not know the problem structure, McClean et al. (2018)
show that

*"for a wide class of reasonable parameterized quantum
circuits, the probability that the gradient along any
reasonable direction is non-zero to some fixed precision
is exponentially small as a function of the number
of qubits."*

Thus, randomly selected quantum circuits might not be the best
option to choose while implementing variational quantum
algorithms.


.. figure:: ../../examples/figures/barren_plateaus/surface.png
   :width: 90%
   :align: center
   :alt: surface

|

In this tutorial, we will show how randomized quantum circuits
face the problem of barren plateaus using PennyLane. We will
partly reproduce some of the findings in McClean et. al., 2018
with just a few lines of code.

.. note::

    **An initialization strategy to tackle barren plateaus**

    How do we avoid the problem of barren plateaus?
    In Grant et al. (2019), the authors present one strategy to
    tackle the barren plateau problem in randomized quantum circuits:

    *"The technique involves randomly selecting some of the initial
    parameter values, then choosing the remaining values so that
    the final circuit is a sequence of shallow unitary blocks that
    each evaluates to the identity. Initializing in this way limits
    the effective depth of the circuits used to calculate the first
    parameter update so that they cannot be stuck in a barren plateau
    at the start of training."*
"""

##############################################################################
# Exploring the barren plateau problem with PennyLane
# ---------------------------------------------------
#
# First, we import PennyLane, NumPy, and Matplotlib

import pennylane as qml
from pennylane import numpy as np
import matplotlib.pyplot as plt


##################################################
# Next, we create a randomized variational circuit

# Set a seed for reproducibility
np.random.seed(20)

num_qubits = 4
dev = qml.device("default.qubit", wires=num_qubits)
gate_set = [qml.RX, qml.RY, qml.RZ]
gate_sequence = {i: np.random.choice(gate_set) for i in range(num_qubits)}


def rand_circuit(params, random_gate_sequence=None, num_qubits=None):
    """A random variational quantum circuit.

    Args:
        params (array[float]): array of parameters
        random_gate_sequence (dict): a dictionary of random gates
        num_qubits (int): the number of qubits in the circuit

    Returns:
        float: the expectation value of the target observable
    """
    for i in range(num_qubits):
        qml.RY(np.pi / 4, wires=i)

    for i in range(num_qubits):
        random_gate_sequence[i](params[i], wires=i)

    for i in range(num_qubits - 1):
        qml.CZ(wires=[i, i + 1])

    H = np.zeros((2 ** num_qubits, 2 ** num_qubits))
    H[0, 0] = 1
    wirelist = [i for i in range(num_qubits)]
    return qml.expval(qml.Hermitian(H, wirelist))


############################################################
# Now we can compute the gradient and calculate the variance.
# While we only take 200 samples to allow the code to run in
# a reasonable amount of time, this can be increased
# for more accurate results.

qcircuit = qml.QNode(rand_circuit, dev)
grad = qml.grad(qcircuit, argnum=0)

num_samples = 200
grad_vals = []

for i in range(num_samples):
    params = np.random.uniform(0, 2 * np.pi, size=num_qubits)
    grad_vals.append(grad(params, random_gate_sequence=gate_sequence, num_qubits=num_qubits))

print("Variance of the gradient for {} samples: {}".format(num_samples, np.var(grad_vals)))


###########################################################
# Evaluate the gradient for more qubits
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# We can repeat the above analysis with increasing number
# of qubits.


def generate_random_circuit(num_qubits):
    """
    Generates a random quantum circuit based on (McClean et. al., 2019).

    Args:
        num_qubits (int): the number of qubits in the circuit
    """
    dev = qml.device("default.qubit", wires=num_qubits)
    gate_set = [qml.RX, qml.RY, qml.RZ]
    random_gate_sequence = {i: np.random.choice(gate_set) for i in range(num_qubits)}
    qcircuit = qml.QNode(rand_circuit, dev)

    return qcircuit, random_gate_sequence


qubits = [2, 3, 4, 5, 6]
variances = []

num_samples = 200


for num_qubits in qubits:
    qcircuit, gate_sequence = generate_random_circuit(num_qubits)
    grad = qml.grad(qcircuit, argnum=0)
    grad_vals = []
    for i in range(num_samples):
        params = np.random.uniform(0, 2 * np.pi, size=num_qubits)
        grad_vals.append(grad(params, random_gate_sequence=gate_sequence, num_qubits=num_qubits))
    vargrad = np.var(grad_vals)
    variances.append(vargrad)
    print("Variance of the gradient for {} qubits: {}".format(num_qubits, vargrad))
    

variances = np.array(variances)
qubits = np.array(qubits)


# Fit the semilog plot to a straight line
p = np.polyfit(qubits, np.log(variances), 1)


# Plot the straight line fit to the semilog
plt.semilogy(qubits, variances, "o")
plt.semilogy(qubits, np.exp(p[0] * qubits + p[1]), "o-.",
             label="Slope {:3.2f}".format(p[0]))
plt.xlabel(r"N Qubits")
plt.ylabel(r"$\langle \partial \theta_{1, 1} E\rangle$ variance")
plt.legend()
plt.show()



##############################################################################
# This tutorial was generated using the following PennyLane version:

qml.about()


##############################################################################
# References
# ----------
#
# 1. Dauphin, Yann N., et al.,
#    Identifying and attacking the saddle point problem in high-dimensional non-convex
#    optimization. Advances in Neural Information Processing
#    systems (2014).
#
# 2. McClean, Jarrod R., et al.,
#    Barren plateaus in quantum neural network training landscapes.
#    Nature communications 9.1 (2018): 4812.
#
# 3. Grant, Edward, et al.
#    An initialization strategy for addressing barren plateaus in
#    parametrized quantum circuits. arXiv preprint arXiv:1903.05076 (2019).



r"""

.. _advanced_features:

Advanced Usage
==============

In the previous three introductory tutorials (:ref:`qubit rotation <qubit_rotation>`,
:ref:`Gaussian transformation <gaussian_transformation>`, and
:ref:`plugins & hybrid computation <plugins_hybrid>`) we explored the basic concepts of
PennyLane, including qubit- and CV-model quantum computations, gradient-based optimization,
and the construction of hybrid classical-quantum computations.

In this tutorial, we will highlight some of the more advanced features of Pennylane.
"""

##############################################################################
# Multiple measurements
# ---------------------
#
# In all the previous examples, we considered quantum functions with only single expectation values.
# In fact, PennyLane supports the return of multiple measurements, up to one per wire.
#
# As usual, we begin by importing PennyLane and the PennyLane-provided version of NumPy, and
# set up a 2-wire qubit device for computations:

import pennylane as qml
from pennylane import numpy as np

dev = qml.device("default.qubit", wires=2)

##############################################################################
# We will start with a simple example circuit, which generates a two-qubit entangled state,
# then evaluates the expectation value of the Pauli Z operator on each wire.


@qml.qnode(dev)
def circuit1(param):
    qml.RX(param, wires=0)
    qml.CNOT(wires=[0, 1])
    return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliZ(1))


##############################################################################
# The degree of entanglement of the qubits is determined by the value of ``param``. For a value of
# :math:`\frac{\pi}{2}`, they are maximally entangled. In this case, the reduced states on each
# subsystem are completely mixed, and local expectation values — like those we are measuring —
# will average to zero.

print(circuit1(np.pi / 2))

##############################################################################
# Notice that the output of the circuit is a NumPy array with ``shape=(2,)``, i.e., a two-dimensional
# vector. These two dimensions match the number of expectation values returned in our quantum
# function ``circuit1``.
#
# .. note::
#
#     It is important to emphasize that the expectation values in ``circuit`` are both **local**,
#     i.e., this circuit is evaluating :math:`\left\langle \sigma_z\right\rangle_0` and :math:`\left\langle \sigma_z\right\rangle_1`,
#     not :math:`\left\langle \sigma_z\otimes \sigma_z\right\rangle_{01}` (where the subscript denotes which wires the
#     observable is located on).
#
# We may even mix different return types, for example expectation values and variances:


@qml.qnode(dev)
def circuit1(param):
    qml.RX(param, wires=0)
    qml.CNOT(wires=[0, 1])
    return qml.expval(qml.PauliZ(0)), qml.var(qml.PauliZ(1))


##############################################################################
# Keyword arguments
# -----------------
#
# While automatic differentiation is a handy feature, sometimes we want certain parts of our
# computational pipeline (e.g., the inputs :math:`x` to a parameterized quantum function
# :math:`f(x;\bf{\theta})` or the training data for a machine learning model) to not be
# differentiated.
#
# PennyLane uses the pattern that *all positional arguments to quantum functions are available
# to be differentiated*, while *keyword arguments are never differentiated*. Thus, when using the
# gradient-descent-based optimizers included in PennyLane, all
# numerical parameters appearing in non-keyword arguments will be updated, while all numerical
# values included as keyword arguments will not be updated.
#
# .. note::
#
#     When constructing the circuit, keyword arguments are defined by providing a
#     **default value** in the function signature. If you would prefer that the keyword argument
#     value be passed every time the quantum circuit function is called, the default value
#     can be set to ``None``.
#
# For example, let's create a quantum node that accepts two arguments; a differentiable
# circuit parameter ``param``, and a fixed circuit parameter ``fixed``:


@qml.qnode(dev)
def circuit3(param, fixed=None):
    qml.RX(fixed, wires=0)
    qml.RX(param, wires=1)
    qml.CNOT(wires=[0, 1])
    return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliZ(1))


##############################################################################
# Calling the circuit, we can feed values to the keyword argument ``fixed``:

print(circuit3(0.1, fixed=-0.2))

print(circuit3(0.1, fixed=1.2))

##############################################################################
# Since keyword arguments do not get considered when computing gradients, the
# Jacobian will still be a 2-dimensional vector.

j3 = qml.jacobian(circuit3, argnum=0)
print(j3(2.5, fixed=3.2))

##############################################################################
# Once defined, keyword arguments must *always* be passed as keyword arguments. PennyLane does
# not support passing keyword argument values as positional arguments.
#
# For example, the following circuit evaluation will correctly update the value of the fixed parameter:

print(circuit3(0.1, fixed=0.4))

##############################################################################
# However, attempting to pass the fixed parameter as a positional argument will
# not work, and PennyLane will attempt to use the default value (``None``) instead:
#
# >>> circuit3(0.1, 0.4)
# ---------------------------------------------------------------------------
# TypeError                                 Traceback (most recent call last)
# <ipython-input-6-949e31911afa> in <module>()
# ----> 1 circuit3(0.1, 0.4)
# ~/pennylane/variable.py in val(self)
#     134
#     135         # The variable is a placeholder for a keyword argument
# --> 136         value = self.kwarg_values[self.name][self.idx] * self.mult
#     137         return value
# TypeError: unsupported operand type(s) for *: 'NoneType' and 'int'



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for the :mod:`pennylane.qnode` decorator.
"""
# pylint: disable=protected-access,cell-var-from-loop
import numpy as np

import pennylane as qml


class TestMethodBinding:
    """Test QNode methods are correctly bound to
    the wrapped function"""

    def test_jacobian(self):
        """Test binding of jacobian method"""
        dev = qml.device('default.qubit', wires=1)

        @qml.qnode(dev)
        def circuit(a):
            qml.RX(a, wires=0)
            return qml.expval(qml.PauliZ(wires=0))

        assert hasattr(circuit, 'jacobian')

        def circuit2(a):
            qml.RX(a, wires=0)
            return qml.expval(qml.PauliZ(wires=0))

        qnode = qml.QNode(circuit2, dev)

        assert qnode.jacobian(0.5) == circuit.jacobian(0.5)

    def test_metric_tensor(self, tol):
        """Test binding of metric tensor methods"""
        dev = qml.device('default.qubit', wires=1)

        a, b = 0.4, 0.1

        @qml.qnode(dev)
        def circuit(a, b):
            qml.RX(a, wires=0)
            qml.RY(b, wires=0)
            return qml.expval(qml.PauliZ(wires=0))

        assert hasattr(circuit, 'metric_tensor')

        def circuit2(a, b):
            qml.RX(a, wires=0)
            qml.RY(b, wires=0)
            return qml.expval(qml.PauliZ(wires=0))

        qnode = qml.QNode(circuit2, dev)

        # check that both QNode constructions agree
        res = circuit.metric_tensor(a, b)
        res2 = qnode.metric_tensor(a, b)
        assert np.allclose(res, res2, atol=tol, rtol=0)

        # check metric tensor is correct
        expected = np.diag(np.array([1, np.cos(a)**2])/4)
        assert np.allclose(res, expected, atol=tol, rtol=0)



"""
.. _qaoa_maxcut:

QAOA for MaxCut
===============

In this tutorial we implement the quantum approximate optimization algorithm (QAOA) for the MaxCut
problem as proposed by `Farhi, Goldstone, and Gutmann (2014) <https://arxiv.org/abs/1411.4028>`__. First, we
give an overview of the MaxCut problem using a simple example, a graph with 4 vertices and 4 edges. We then
show how to find the maximum cut by running the QAOA algorithm using PennyLane.
"""

##############################################################################
# Background
# ----------
#
# The MaxCut problem
# ~~~~~~~~~~~~~~~~~~
# The aim of MaxCut is to maximize the number of edges (yellow lines) in a graph that are "cut" by
# a given partition of the vertices (blue circles) into two sets (see figure below).
#
# .. figure:: ../../examples/figures/qaoa_maxcut_partition.png
#    :align: center
#    :scale: 65%
#    :alt: qaoa_operators
#
# |
# Consider a graph with :math:`m` edges and :math:`n` vertices. We seek the partition
# :math:`z` of the vertices into two sets
# :math:`A` and :math:`B` which maximizes
#
# .. math::
#   C(z) = \sum_{\alpha=1}^{m}C_\alpha(z),
#
# where :math:`C` counts the number of edges cut. :math:`C_\alpha(z)=1` if :math:`z` places one vertex from the
# :math:`\alpha^\text{th}` edge in set :math:`A` and the other in set :math:`B`, and :math:`C_\alpha(z)=0` otherwise.
# Finding a cut which yields the maximum possible value of :math:`C` is an NP-complete problem, so our best hope for a
# polynomial-time algorithm lies in an approximate optimization.
# In the case of MaxCut, this means finding a partition :math:`z` which
# yields a value for :math:`C(z)` that is close to the maximum possible value.
#
# We can represent the assignment of vertices to set :math:`A` or :math:`B` using a bitstring,
# :math:`z=z_1...z_n` where :math:`z_i=0` if the :math:`i^\text{th}` vertex is in :math:`A` and
# :math:`z_i = 1` if it is in :math:`B`. For instance,
# in the situation depicted in the figure above the bitstring representation is :math:`z=0101\text{,}`
# indicating that the :math:`0^{\text{th}}` and :math:`2^{\text{nd}}` vertices are in :math:`A`
# while the :math:`1^{\text{st}}` and :math:`3^{\text{rd}}` are in
# :math:`B`. This assignment yields a value for the objective function (the number of yellow lines cut)
# :math:`C=4`, which turns out to be the maximum cut. In the following sections,
# we will represent partitions using computational basis states and use PennyLane to
# rediscover this maximum cut.
#
# .. note:: In the graph above, :math:`z=1010` could equally well serve as the maximum cut.
#
# A circuit for QAOA
# ~~~~~~~~~~~~~~~~~~~~
# This section describes implementing a circuit for QAOA using basic unitary gates to find approximate
# solutions to the MaxCut problem.
# Firstly, denoting the partitions using computational basis states :math:`|z\rangle`, we can represent the terms in the
# objective function as operators acting on these states
#
# .. math::
#   C_\alpha = \frac{1}{2}\left(1-\sigma_{z}^j\sigma_{z}^k\right),
#
# where the :math:`\alpha\text{th}` edge is between vertices :math:`(j,k)`.
# :math:`C_\alpha` has eigenvalue 1 if and only if the :math:`j\text{th}` and :math:`k\text{th}`
# qubits have different z-axis measurement values, representing separate partitions.
# The objective function :math:`C` can be considered a diagonal operator with integer eigenvalues.
#
# QAOA starts with a uniform superposition over the :math:`n` bitstring basis states,
#
# .. math::
#   |+_{n}\rangle = \frac{1}{\sqrt{2^n}}\sum_{z\in \{0,1\}^n} |z\rangle.
#
#
# We aim to explore the space of bitstring states for a superposition which is likely to yield a
# large value for the :math:`C` operator upon performing a measurement in the computational basis.
# Using the :math:`2p` angle parameters
# :math:`\boldsymbol{\gamma} = \gamma_1\gamma_2...\gamma_p`, :math:`\boldsymbol{\beta} = \beta_1\beta_2...\beta_p`
# we perform a sequence of operations on our initial state:
#
# .. math::
#   |\boldsymbol{\gamma},\boldsymbol{\beta}\rangle = U_{B_p}U_{C_p}U_{B_{p-1}}U_{C_{p-1}}...U_{B_1}U_{C_1}|+_n\rangle
#
# where the operators have the explicit forms
#
# .. math::
#   U_{B_l} &= e^{-i\beta_lB} = \prod_{j=1}^n e^{-i\beta_l\sigma_x^j}, \\
#   U_{C_l} &= e^{-i\gamma_lC} = \prod_{\text{edge (j,k)}} e^{-i\gamma_l(1-\sigma_z^j\sigma_z^k)/2}.
#
# In other words, we make :math:`p` layers of parametrized :math:`U_bU_C` gates.
# These can be implemented on a quantum circuit using the gates depicted below, up to an irrelevant constant
# that gets absorbed into the parameters.
#
# .. figure:: ../../examples/figures/qaoa_operators.png
#    :align: center
#    :scale: 100%
#    :alt: qaoa_operators
#
# |
# Let :math:`\langle \boldsymbol{\gamma},
# \boldsymbol{\beta} | C | \boldsymbol{\gamma},\boldsymbol{\beta} \rangle` be the expectation of the objective operator.
# In the next section, we will use PennyLane to perform classical optimization
# over the circuit parameters :math:`(\boldsymbol{\gamma}, \boldsymbol{\beta})`.
# This will specify a state :math:`|\boldsymbol{\gamma},\boldsymbol{\beta}\rangle` which is
# likely to yield an approximately optimal partition :math:`|z\rangle` upon performing a measurement in the
# computational basis.
# In the case of the graph shown above, we want to measure either 0101 or 1010 from our state since these correspond to
# the optimal partitions.
#
# .. figure:: ../../examples/figures/qaoa_optimal_state.png
#   :align: center
#   :scale: 60%
#   :alt: optimal_state
#
# |
# Qualitatively, QAOA tries to evolve the initial state into the plane of the
# :math:`|0101\rangle`, :math:`|1010\rangle` basis states (see figure above).
#
#
# Implementing QAOA in PennyLane
# ------------------------------
#
# Imports and setup
# ~~~~~~~~~~~~~~~~~
#
# To get started, we import PennyLane along with the PennyLane-provided
# version of NumPy.


import pennylane as qml
from pennylane import numpy as np


##############################################################################
# Operators
# ~~~~~~~~~
# We specify the number of qubits (vertices) with ``n_wires`` and
# compose the unitary operators using the definitions
# above. :math:`U_B` operators act on individual wires, while :math:`U_C`
# operators act on wires whose corresponding vertices are joined by an edge in
# the graph. We also define the graph using
# the list ``graph``, which contains the tuples of vertices defining
# each edge in the graph.

n_wires = 4
graph = [(0, 1), (0, 3), (1, 2), (2, 3)]

# unitary operator U_B with parameter beta
def U_B(beta):
    for wire in range(n_wires):
        qml.RX(2 * beta, wires=wire)


# unitary operator U_C with parameter gamma
def U_C(gamma):
    for edge in graph:
        wire1 = edge[0]
        wire2 = edge[1]
        qml.CNOT(wires=[wire1, wire2])
        qml.RZ(gamma, wires=wire2)
        qml.CNOT(wires=[wire1, wire2])


##############################################################################
# We will need a way to sample
# a measurement of multiple qubits in the computational basis, so we define
# a Hermitian operator to do this. The eigenvalues of the operator are
# the qubit measurement values in integer form.


def comp_basis_measurement(wires):
    n_wires = len(wires)
    return qml.Hermitian(np.diag(range(2 ** n_wires)), wires=wires)


##############################################################################
# Circuit
# ~~~~~~~
# Next, we create a quantum device with 4 qubits.

dev = qml.device("default.qubit", wires=n_wires, analytic=True, shots=1)

##############################################################################
# We also require a quantum node which will apply the operators according to the
# angle parameters, and return the expectation value of the observable
# :math:`\sigma_z^{j}\sigma_z^{k}` to be used in each term of the objective function later on. The
# argument ``edge`` specifies the chosen edge term in the objective function, :math:`(j,k)`.
# Once optimized, the same quantum node can be used for sampling an approximately optimal bitstring
# if executed with the ``edge`` keyword set to ``None``. Additionally, we specify the number of layers
# (repeated applications of :math:`U_BU_C`) using the keyword ``n_layers``.

pauli_z = [[1, 0], [0, -1]]
pauli_z_2 = np.kron(pauli_z, pauli_z)


@qml.qnode(dev)
def circuit(gammas, betas, edge=None, n_layers=1):
    # apply Hadamards to get the n qubit |+> state
    for wire in range(n_wires):
        qml.Hadamard(wires=wire)
    # p instances of unitary operators
    for i in range(n_layers):
        U_C(gammas[i])
        U_B(betas[i])
    if edge is None:
        # measurement phase
        return qml.sample(comp_basis_measurement(range(n_wires)))
    # during the optimization phase we are evaluating a term
    # in the objective using expval
    return qml.expval(qml.Hermitian(pauli_z_2, wires=edge))


##############################################################################
# Optimization
# ~~~~~~~~~~~~
# Finally, we optimize the objective over the
# angle parameters :math:`\boldsymbol{\gamma}` (``params[0]``) and :math:`\boldsymbol{\beta}`
# (``params[1]``)
# and then sample the optimized
# circuit multiple times to yield a distribution of bitstrings. One of the optimal partitions
# (:math:`z=0101` or :math:`z=1010`) should be the most frequently sampled bitstring.
# We perform a maximization of :math:`C` by
# minimizing :math:`-C`, following the convention that optimizations are cast as minimizations
# in PennyLane.


def qaoa_maxcut(n_layers=1):
    print("\np={:d}".format(n_layers))

    # initialize the parameters near zero
    init_params = 0.01 * np.random.rand(2, n_layers)

    # minimize the negative of the objective function
    def objective(params):
        gammas = params[0]
        betas = params[1]
        neg_obj = 0
        for edge in graph:
            # objective for the MaxCut problem
            neg_obj -= 0.5 * (1 - circuit(gammas, betas, edge=edge, n_layers=n_layers))
        return neg_obj

    # initialize optimizer: Adagrad works well empirically
    opt = qml.AdagradOptimizer(stepsize=0.5)

    # optimize parameters in objective
    params = init_params
    steps = 30
    for i in range(steps):
        params = opt.step(objective, params)
        if (i + 1) % 5 == 0:
            print("Objective after step {:5d}: {: .7f}".format(i + 1, -objective(params)))

    # sample measured bitstrings 100 times
    bit_strings = []
    n_samples = 100
    for i in range(0, n_samples):
        bit_strings.append(int(circuit(params[0], params[1], edge=None, n_layers=n_layers)))

    # print optimal parameters and most frequently sampled bitstring
    counts = np.bincount(np.array(bit_strings))
    most_freq_bit_string = np.argmax(counts)
    print("Optimized (gamma, beta) vectors:\n{}".format(params))
    print("Most frequently sampled bit string is: {:04b}".format(most_freq_bit_string))

    return -objective(params), bit_strings


# perform qaoa on our graph with p=1,2 and
# keep the bitstring sample lists
bitstrings1 = qaoa_maxcut(n_layers=1)[1]
bitstrings2 = qaoa_maxcut(n_layers=2)[1]

##############################################################################
# In the case where we set ``n_layers=2``, we recover the optimal
# objective function :math:`C=4`

##############################################################################
# Plotting the results
# --------------------
# We can plot the distribution of measurements obtained from the optimized circuits. As
# expected for this graph, the partitions 0101 and 1010 are measured with the highest frequencies,
# and in the case where we set ``n_layers=2`` we obtain one of the optimal partitions with 100% certainty.

import matplotlib.pyplot as plt

xticks = range(0, 16)
xtick_labels = list(map(lambda x: format(x, "04b"), xticks))
bins = np.arange(0, 17) - 0.5

fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(8, 4))
plt.subplot(1, 2, 1)
plt.title("n_layers=1")
plt.xlabel("bitstrings")
plt.ylabel("freq.")
plt.xticks(xticks, xtick_labels, rotation="vertical")
plt.hist(bitstrings1, bins=bins)
plt.subplot(1, 2, 2)
plt.title("n_layers=2")
plt.xlabel("bitstrings")
plt.ylabel("freq.")
plt.xticks(xticks, xtick_labels, rotation="vertical")
plt.hist(bitstrings2, bins=bins)
plt.tight_layout()
plt.show()



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
This module contains functions that generate initial parameters, for example
to use in templates.
"""
# pylint: disable=too-many-arguments
from math import pi
import numpy as np


def strong_ent_layers_uniform(n_layers, n_wires, low=0, high=2 * pi, seed=None):
    r"""Creates a list of a single parameter array for :func:`~.StronglyEntanglingLayers`, drawn from a uniform
    distribution.

    The shape of the parameter array is ``(n_layers, n_wires, 3)`` and each parameter is drawn uniformly at random \
    from between ``low`` and ``high``. The parameters define the three rotation angles
    applied in each layer.

    Args:
        n_layers (int): number of layers
        n_wires (int): number of qubits

    Keyword Args:
        low (float): minimum value of non-angle gate parameters
        high (float): maximum value of non-angle gate parameters
        seed (int): seed used in sampling the parameters, makes function call deterministic

    Returns:
        list of parameter array
    """
    if seed is not None:
        np.random.seed(seed)

    params = np.random.uniform(low=low, high=high, size=(n_layers, n_wires, 3))
    return [params]


def strong_ent_layers_normal(n_layers, n_wires, mean=0, std=0.1, seed=None):
    r"""Creates a list of a single parameter array for :func:`~.StronglyEntanglingLayers`, drawn from a normal
    distribution.

    The shape of the parameter array is ``(n_layers, n_wires, 3)`` and each parameter is drawn
    from a normal distribution with mean ``mean`` and standard deviation ``std``.
    The parameters define the three rotation angles applied in each layer.

    Args:
        n_layers (int): number of layers
        n_wires (int): number of qubits

    Keyword Args:
        mean (float): mean of parameters
        std (float): standard deviation of parameters
        seed (int): seed used in sampling the parameters, makes function call deterministic

    Returns:
        list of parameter array
    """
    if seed is not None:
        np.random.seed(seed)

    params = np.random.normal(loc=mean, scale=std, size=(n_layers, n_wires, 3))
    return [params]


def strong_ent_layer_uniform(n_wires, low=0, high=2 * pi, seed=None):
    r"""Creates a list of a single parameter array for :func:`~.StronglyEntanglingLayer`,
    drawn from a uniform distribution.

    The shape of the parameter array is ``(n_wires, 3)`` and each parameter is drawn uniformly at random \
    from between ``low`` and ``high``. The parameters define the three rotation angles
    applied to each layer.

    Args:
        n_wires (int): number of qubits

    Keyword Args:
        low (float): minimum value of non-angle gate parameters
        high (float): maximum value of non-angle gate parameters
        seed (int): seed used in sampling the parameters, makes function call deterministic

    Returns:
        list of parameter array
    """
    if seed is not None:
        np.random.seed(seed)

    params = np.random.uniform(low=low, high=high, size=(n_wires, 3))
    return [params]


def strong_ent_layer_normal(n_wires, mean=0, std=0.1, seed=None):
    r"""Creates a list of a single parameter array for :func:`~.StronglyEntanglingLayer`,
    drawn from a normal distribution.

    The shape of the parameter array is ``(n_wires, 3)`` and each parameter is drawn
    from a normal distribution with mean ``mean`` and standard deviation ``std``.
    The parameters define the three rotation angles applied to each layer.

    Args:
        n_wires (int): number of qubits

    Keyword Args:
        mean (float): mean of parameters
        std (float): standard deviation of parameters
        seed (int): seed used in sampling the parameters, makes function call deterministic

    Returns:
        list of parameter array
    """
    if seed is not None:
        np.random.seed(seed)

    params = np.random.normal(loc=mean, scale=std, size=(n_wires, 3))
    return [params]


def random_layers_uniform(n_layers, n_wires, n_rots=None, low=0, high=2 * pi, seed=None):
    r"""Creates a list of a single parameter array for :func:`~.RandomLayers`, drawn from a uniform distribution.

    The shape of the parameter array is ``(n_layers, n_rots)`` and each parameter is drawn uniformly at random \
    from between ``low`` and ``high``. The parameters define the rotation angles of the randomly \
    positioned rotations applied in each layer.

    Args:
        n_layers (int): number of layers
        n_wires (int): number of qubits

    Keyword Args:
        n_rots (int): number of rotations, if ``None``, ``n_rots=n_wires``
        low (float): minimum value of non-angle gate parameters
        high (float): maximum value of non-angle gate parameters
        seed (int): seed used in sampling the parameters, makes function call deterministic

    Returns:
        list of parameter array
    """
    if seed is not None:
        np.random.seed(seed)

    if n_rots is None:
        n_rots = n_wires

    params = np.random.uniform(low=low, high=high, size=(n_layers, n_rots))
    return [params]


def random_layers_normal(n_layers, n_wires, n_rots=None, mean=0, std=0.1, seed=None):
    r"""Creates a list of a single parameter array for :func:`~.RandomLayers`, drawn from a normal distribution.

    The shape of the parameter array is ``(n_layers, n_rots)`` and each parameter is drawn
    from a normal distribution with mean ``mean`` and standard deviation ``std``.
    The parameters define the rotation angles of the randomly positioned rotations applied in each layer.

    Args:
        n_layers (int): number of layers
        n_wires (int): number of qubits

    Keyword Args:
        n_rots (int): number of rotations, if ``None``, ``n_rots=n_wires``
        mean (float): mean of parameters
        std (float): standard deviation of parameters
        seed (int): seed used in sampling the parameters, makes function call deterministic

    Returns:
        list of parameter array
    """
    if seed is not None:
        np.random.seed(seed)

    if n_rots is None:
        n_rots = n_wires

    params = np.random.normal(loc=mean, scale=std, size=(n_layers, n_rots))
    return [params]


def random_layer_uniform(n_wires, n_rots=None, low=0, high=2 * pi, seed=None):
    r"""Creates a list of a single parameter array for :func:`~.RandomLayer`, drawn from a uniform distribution.

    The number of parameter array is ``(n_rots,)`` and each parameter is drawn uniformly at random \
    from between ``low`` and ``high``. The parameters define the rotation angles of the randomly \
    positioned rotations applied in each layer.

    Args:
        n_wires (int): number of qubits

    Keyword Args:
        n_rots (int): number of rotations, if ``None``, ``n_rots=n_wires``
        low (float): minimum value of rotation angles
        high (float): maximum value of rotation angles
        seed (int): seed used in sampling the parameters, makes function call deterministic

    Returns:
        list of parameter array
    """
    if seed is not None:
        np.random.seed(seed)

    if n_rots is None:
        n_rots = n_wires

    params = np.random.uniform(low=low, high=high, size=(n_rots,))
    return [params]


def random_layer_normal(n_wires, n_rots=None, mean=0, std=0.1, seed=None):
    r"""Creates a list of a single parameter array for :func:`~.RandomLayer`, drawn from a normal distribution.

    The number of parameter array is ``(n_rots,)`` and each parameter is drawn
    from a normal distribution with mean ``mean`` and standard deviation ``std``.
    The parameters define the rotation angles of the randomly positioned rotations applied in each layer.

    Args:
        n_wires (int): number of qubits

    Keyword Args:
        n_rots (int): number of rotations, if ``None``, ``n_rots=n_wires``
        mean (float): mean of parameters
        std (float): standard deviation of parameters
        seed (int): seed used in sampling the parameters, makes function call deterministic

    Returns:
        list of parameter array
    """
    if seed is not None:
        np.random.seed(seed)

    if n_rots is None:
        n_rots = n_wires

    params = np.random.normal(loc=mean, scale=std, size=(n_rots,))
    return [params]


def cvqnn_layers_uniform(n_layers, n_wires, low=0, high=2*pi, mean_active=0, std_active=0.1, seed=None):
    r"""Creates a list of eleven parameter arrays for :func:`~.CVNeuralNetLayers`,
    where non-active gate parameters are drawn from a uniform distribution and active parameters
    from a normal distribution.

    The shape of the arrays is ``(n_layers, n_wires*(n_wires-1)/2)`` for the parameters used in an interferometer,
    and ``(n_layers, n_wires)``  else.

    All gate parameters are drawn uniformly from the interval ``[low, high]``, except from the three types of
    'active gate parameters': the displacement amplitude, squeezing amplitude and kerr parameter. These
    active gate parameters are sampled from a normal distribution with mean ``mean_active`` and standard
    deviation ``std_active``. Since they influence the mean photon number (or energy) of the quantum system,
    one typically wants to initialize them with values close to zero.

    Args:
        n_layers (int): number of layers of the CV Neural Net
        n_wires (int): number of modes of the CV Neural Net

    Keyword Args:
        low (float): minimum value of uniformly drawn rotation angles
        high (float): maximum value of uniformly drawn rotation angles
        mean_active (float): mean of active gate parameters
        std_active (float): standard deviation of active gate parameters
        seed (int): seed used in sampling the parameters, makes function call deterministic

    Returns:
        list of parameter arrays
    """
    if seed is not None:
        np.random.seed(seed)
    n_if = n_wires * (n_wires - 1) // 2

    theta_1 = np.random.uniform(low=low, high=high, size=(n_layers, n_if))
    phi_1 = np.random.uniform(low=low, high=high, size=(n_layers, n_if))
    varphi_1 = np.random.uniform(low=low, high=high, size=(n_layers, n_wires))
    r = np.random.normal(loc=mean_active, scale=std_active, size=(n_layers, n_wires))
    phi_r = np.random.uniform(low=low, high=high, size=(n_layers, n_wires))
    theta_2 = np.random.uniform(low=low, high=high, size=(n_layers, n_if))
    phi_2 = np.random.uniform(low=low, high=high, size=(n_layers, n_if))
    varphi_2 = np.random.uniform(low=low, high=high, size=(n_layers, n_wires))
    a = np.random.normal(loc=mean_active, scale=std_active, size=(n_layers, n_wires))
    phi_a = np.random.uniform(low=low, high=high, size=(n_layers, n_wires))
    k = np.random.normal(loc=mean_active, scale=std_active, size=(n_layers, n_wires))

    return [theta_1, phi_1, varphi_1, r, phi_r, theta_2, phi_2, varphi_2, a, phi_a, k]


def cvqnn_layers_normal(n_layers, n_wires, mean=0, std=1, mean_active=0, std_active=0.1, seed=None):
    r"""Creates a list of eleven parameter arrays for :func:`~.CVNeuralNetLayers`, where both active and non-active
    gate parameters are drawn from normal distributions.

    The shape of the arrays is ``(n_layers, n_wires*(n_wires-1)/2)`` for the parameters used in an interferometer,
    and ``(n_layers, n_wires)``  else.

    All gate parameters are drawn from a normal distribution with mean ``mean`` and standard deviation ``std``,
    except from the three types of 'active gate parameters': the displacement amplitude, squeezing amplitude and kerr
    parameter. These active gate parameters are sampled from a normal distribution with mean ``mean_active`` and
    standard deviation ``std_active``. Since they influence the mean photon number (or energy) of the quantum system,
    one typically wants to initialize them with values close to zero.

    Args:
        n_layers (int): number of layers of the CV Neural Net
        n_wires (int): number of modes of the CV Neural Net

    Keyword Args:
        mean (float): mean of non-active parameters
        std (float): standard deviation of non-active parameters
        mean_active (float): mean of active gate parameters
        std_active (float): standard deviation of active gate parameters
        seed (int): seed used in sampling the parameters, makes function call deterministic

    Returns:
        list of parameter arrays
    """
    if seed is not None:
        np.random.seed(seed)
    n_if = n_wires * (n_wires - 1) // 2

    theta_1 = np.random.normal(loc=mean, scale=std, size=(n_layers, n_if))
    phi_1 = np.random.normal(loc=mean, scale=std, size=(n_layers, n_if))
    varphi_1 = np.random.normal(loc=mean, scale=std, size=(n_layers, n_wires))
    r = np.random.normal(loc=mean_active, scale=std_active, size=(n_layers, n_wires))
    phi_r = np.random.normal(loc=mean, scale=std, size=(n_layers, n_wires))
    theta_2 = np.random.normal(loc=mean, scale=std, size=(n_layers, n_if))
    phi_2 = np.random.normal(loc=mean, scale=std, size=(n_layers, n_if))
    varphi_2 = np.random.normal(loc=mean, scale=std, size=(n_layers, n_wires))
    a = np.random.normal(loc=mean_active, scale=std_active, size=(n_layers, n_wires))
    phi_a = np.random.normal(loc=mean, scale=std, size=(n_layers, n_wires))
    k = np.random.normal(loc=mean_active, scale=std_active, size=(n_layers, n_wires))

    return [theta_1, phi_1, varphi_1, r, phi_r, theta_2, phi_2, varphi_2, a, phi_a, k]


def cvqnn_layer_uniform(n_wires, low=0, high=2 * pi, mean_active=0, std_active=0.1, seed=None):
    r"""Creates a list of eleven parameter arrays for :func:`~.CVNeuralNetLayer`,
    where non-active gate parameters are drawn from a uniform distribution and active parameters
    from a normal distribution.

    The shape of the arrays is ``(n_wires*(n_wires-1)/2)`` for the parameters used in an interferometer,
    and ``(n_wires)``  else.

    All gate parameters are drawn uniformly from the interval ``[low, high]``, except from the three types of
    'active gate parameters': the displacement amplitude, squeezing amplitude and kerr parameter. These
    active gate parameters are sampled from a normal distribution with mean ``mean_active`` and standard
    deviation ``std_active``. Since they influence the mean photon number (or energy) of the quantum system,
    one typically wants to initialize them with values close to zero.

    Args:
        n_wires (int): number of modes of the CV Neural Net

    Keyword Args:
        low (float): minimum value of uniformly drawn non-active gate parameters
        high (float): maximum value of uniformly drawn non-active gate parameters
        mean_active (float): mean of active gate parameters
        std_active (float): standard deviation of active gate parameters
        seed (int): seed used in sampling the parameters, makes function call deterministic

    Returns:
        list of parameter arrays
    """
    if seed is not None:
        np.random.seed(seed)

    n_if = n_wires * (n_wires - 1) // 2

    theta_1 = np.random.uniform(low=low, high=high, size=(n_if, ))
    phi_1 = np.random.uniform(low=low, high=high, size=(n_if, ))
    varphi_1 = np.random.uniform(low=low, high=high, size=(n_wires,))
    r = np.random.normal(loc=mean_active, scale=std_active, size=(n_wires,))
    phi_r = np.random.uniform(low=low, high=high, size=(n_wires,))
    theta_2 = np.random.uniform(low=low, high=high, size=(n_if, ))
    phi_2 = np.random.uniform(low=low, high=high, size=(n_if, ))
    varphi_2 = np.random.uniform(low=low, high=high, size=(n_wires,))
    a = np.random.normal(loc=mean_active, scale=std_active, size=(n_wires,))
    phi_a = np.random.uniform(low=low, high=high, size=(n_wires,))
    k = np.random.normal(loc=mean_active, scale=std_active, size=(n_wires,))

    return [theta_1, phi_1, varphi_1, r, phi_r, theta_2, phi_2, varphi_2, a, phi_a, k]


def cvqnn_layer_normal(n_wires, mean=0, std=1, mean_active=0, std_active=0.1, seed=None):
    r"""Creates a list of eleven parameter arrays for :func:`~.CVNeuralNetLayer`, where both active and non-active
    gate parameters are drawn from normal distributions.

    The shape of the arrays is ``(n_wires*(n_wires-1)/2)`` for the parameters used in an interferometer,
    and ``(n_wires)``  else.

    All gate parameters are drawn from a normal distribution with mean ``mean`` and standard deviation ``std``,
    except from the three types of 'active gate parameters': the displacement amplitude, squeezing amplitude and kerr
    parameter. These active gate parameters are sampled from a normal distribution with mean ``mean_active`` and
    standard deviation ``std_active``. Since they influence the mean photon number (or energy) of the quantum system,
    one typically wants to initialize them with values close to zero.

    Args:
        n_wires (int): number of modes of the CV Neural Net

    Keyword Args:
        mean (float): mean of non-active parameters
        std (float): standard deviation of non-active parameters
        mean_active (float): mean of active gate parameters
        std_active (float): standard deviation of active gate parameters
        seed (int): seed used in sampling the parameters, makes function call deterministic

    Returns:
        list of parameter arrays
    """
    if seed is not None:
        np.random.seed(seed)

    n_if = n_wires * (n_wires - 1) // 2

    theta_1 = np.random.normal(loc=mean, scale=std, size=(n_if,))
    phi_1 = np.random.normal(loc=mean, scale=std, size=(n_if,))
    varphi_1 = np.random.normal(loc=mean, scale=std, size=(n_wires,))
    r = np.random.normal(loc=mean_active, scale=std_active, size=(n_wires,))
    phi_r = np.random.normal(loc=mean, scale=std, size=(n_wires,))
    theta_2 = np.random.normal(loc=mean, scale=std, size=(n_if,))
    phi_2 = np.random.normal(loc=mean, scale=std, size=(n_if,))
    varphi_2 = np.random.normal(loc=mean, scale=std, size=(n_wires,))
    a = np.random.normal(loc=mean_active, scale=std_active, size=(n_wires,))
    phi_a = np.random.normal(loc=mean, scale=std, size=(n_wires,))
    k = np.random.normal(loc=mean_active, scale=std_active, size=(n_wires,))

    return [theta_1, phi_1, varphi_1, r, phi_r, theta_2, phi_2, varphi_2, a, phi_a, k]


def interferometer_uniform(n_wires, low=0, high=2 * pi, seed=None):
    r"""Creates a list of three parameter arrays for an :func:`Interferometer`, drawn from a uniform
    distribution.

    The arrays are of the form ``[theta, phi, varphi]``, where:

    * ``theta`` is the array of beamsplitter transmittivity angles, of size ``(n_wires*(n_wires-1)/2, )``

    * ``phi`` is the array of beamsplitter phases, of size ``(n_wires*(n_wires-1)/2, )``

    * ``varphi`` is the array of local angles for the final rotation gates, of size ``(n_wires, )``

    All parameters are initialized uniformly from the interval ``[low, high]``.

    Args:
        n_wires (int): number of modes that the interferometer acts on

    Keyword Args:
        low (float): minimum value of uniformly drawn rotation angles
        high (float): maximum value of uniformly drawn rotation angles
        seed (int): seed used in sampling the parameters, makes function call deterministic

    Returns:
        list of parameter arrays
    """
    if seed is not None:
        np.random.seed(seed)
    n_if = n_wires * (n_wires - 1) // 2

    theta = np.random.uniform(low=low, high=high, size=(n_if,))
    phi = np.random.uniform(low=low, high=high, size=(n_if,))
    varphi = np.random.uniform(low=low, high=high, size=(n_wires,))

    return [theta, phi, varphi]


def interferometer_normal(n_wires, mean=0, std=0.1, seed=None):
    r"""Creates a list of three parameter arrays for an :func:`Interferometer`, drawn from a normal
    distribution.

    The arrays are of the form ``[theta, phi, varphi]``, where:


    * ``theta`` is the array of beamsplitter transmittivity angles, of size ``(n_wires*(n_wires-1)/2, )``

    * ``phi`` is the array of beamsplitter phases, of size ``(n_wires*(n_wires-1)/2, )``

    * ``varphi`` is the array of local angles for the final rotation gates, of size ``(n_wires, )``

    All parameters are drawn from a normal distribution with mean ``mean`` and standard deviation ``std``.

    Args:
        n_wires (int): number of modes that the interferometer acts on

    Keyword Args:
        mean (float): mean of parameters
        std (float): standard deviation of parameters
        seed (int): seed used in sampling the parameters, makes function call deterministic

    Returns:
        list of parameter arrays
    """
    if seed is not None:
        np.random.seed(seed)
    n_if = n_wires * (n_wires - 1) // 2

    theta = np.random.normal(loc=mean, scale=std, size=(n_if,))
    phi = np.random.normal(loc=mean, scale=std, size=(n_if,))
    varphi = np.random.normal(loc=mean, scale=std, size=(n_wires,))

    return [theta, phi, varphi]



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for the :mod:`pennylane` :class:`Device` class.
"""

import pytest
import pennylane as qml
from pennylane import Device, DeviceError
from pennylane.operation import Sample, Variance, Expectation
from pennylane.qnode import QuantumFunctionError


@pytest.fixture(scope="function")
def mock_device_with_operations(monkeypatch):
    """A mock instance of the abstract Device class with non-empty operations"""
    with monkeypatch.context() as m:
        m.setattr(Device, '__abstractmethods__', frozenset())
        m.setattr(Device, 'operations', ["PauliX", "PauliZ", "CNOT"])
        yield Device()


@pytest.fixture(scope="function")
def mock_device_with_observables(monkeypatch):
    """A mock instance of the abstract Device class with non-empty observables"""
    with monkeypatch.context() as m:
        m.setattr(Device, '__abstractmethods__', frozenset())
        m.setattr(Device, 'observables', ["PauliX", "PauliZ"])
        yield Device()


mock_device_paulis = ["PauliX", "PauliY", "PauliZ"]

@pytest.fixture(scope="function")
def mock_device_supporting_paulis(monkeypatch):
    """A mock instance of the abstract Device class with non-empty observables"""
    with monkeypatch.context() as m:
        m.setattr(Device, '__abstractmethods__', frozenset())
        m.setattr(Device, 'operations', mock_device_paulis)
        m.setattr(Device, 'observables', mock_device_paulis)
        m.setattr(Device, 'short_name', 'MockDevice')
        yield Device()


mock_device_capabilities = {
    "measurements": "everything",
    "noise_models": ["depolarizing", "bitflip"],
}


@pytest.fixture(scope="function")
def mock_device_with_capabilities(monkeypatch):
    """A mock instance of the abstract Device class with non-empty observables"""
    with monkeypatch.context() as m:
        m.setattr(Device, '__abstractmethods__', frozenset())
        m.setattr(Device, '_capabilities', mock_device_capabilities)
        yield Device()


@pytest.fixture(scope="function")
def mock_device_with_paulis_and_methods(monkeypatch):
    """A mock instance of the abstract Device class with non-empty observables"""
    with monkeypatch.context() as m:
        m.setattr(Device, '__abstractmethods__', frozenset())
        m.setattr(Device, '_capabilities', mock_device_capabilities)
        m.setattr(Device, 'operations', mock_device_paulis)
        m.setattr(Device, 'observables', mock_device_paulis)
        m.setattr(Device, 'short_name', 'MockDevice')
        m.setattr(Device, 'expval', lambda self, x, y, z: 0)
        m.setattr(Device, 'var', lambda self, x, y, z: 0)
        m.setattr(Device, 'sample', lambda self, x, y, z: 0)
        m.setattr(Device, 'apply', lambda self, x, y, z: None)
        yield Device()


@pytest.fixture(scope="function")
def mock_device(monkeypatch):
    with monkeypatch.context() as m:
        m.setattr(Device, '__abstractmethods__', frozenset())
        m.setattr(Device, '_capabilities', mock_device_capabilities)
        m.setattr(Device, 'operations', ["PauliY", "RX", "Rot"])
        m.setattr(Device, 'observables', ["PauliZ"])
        m.setattr(Device, 'short_name', 'MockDevice')
        m.setattr(Device, 'expval', lambda self, x, y, z: 0)
        m.setattr(Device, 'var', lambda self, x, y, z: 0)
        m.setattr(Device, 'sample', lambda self, x, y, z: 0)
        m.setattr(Device, 'apply', lambda self, x, y, z: None)
        yield Device()




class TestDeviceSupportedLogic:
    """Test the logic associated with the supported operations and observables"""

    def test_supports_operation_argument_types(self, mock_device_with_operations):
        """Checks that device.supports_operations returns the correct result
           when passed both string and Operation class arguments"""

        assert mock_device_with_operations.supports_operation("PauliX")
        assert mock_device_with_operations.supports_operation(qml.PauliX)

        assert not mock_device_with_operations.supports_operation("PauliY")
        assert not mock_device_with_operations.supports_operation(qml.PauliY)

    def test_supports_observable_argument_types(self, mock_device_with_observables):
        """Checks that device.supports_observable returns the correct result
           when passed both string and Operation class arguments"""

        assert mock_device_with_observables.supports_observable("PauliX")
        assert mock_device_with_observables.supports_observable(qml.PauliX)

        assert not mock_device_with_observables.supports_observable("PauliY")
        assert not mock_device_with_observables.supports_observable(qml.PauliY)

    def test_supports_operation_exception(self, mock_device):
        """check that device.supports_operation raises proper errors
           if the argument is of the wrong type"""

        with pytest.raises(
            ValueError,
            match="The given operation must either be a pennylane.Operation class or a string.",
        ):
            mock_device.supports_operation(3)

        with pytest.raises(
            ValueError,
            match="The given operation must either be a pennylane.Operation class or a string.",
        ):
            mock_device.supports_operation(Device)

    def test_supports_observable_exception(self, mock_device):
        """check that device.supports_observable raises proper errors
           if the argument is of the wrong type"""

        with pytest.raises(
            ValueError,
            match="The given operation must either be a pennylane.Observable class or a string.",
        ):
            mock_device.supports_observable(3)

        with pytest.raises(
            ValueError,
            match="The given operation must either be a pennylane.Observable class or a string.",
        ):
            mock_device.supports_observable(qml.CNOT)



class TestInternalFunctions:
    """Test the internal functions of the abstract Device class"""

    def test_check_validity_on_valid_queue(self, mock_device_supporting_paulis):
        """Tests the function Device.check_validity with valid queue and observables"""
        queue = [
            qml.PauliX(wires=0),
            qml.PauliY(wires=1),
            qml.PauliZ(wires=2),
        ]

        observables = [qml.expval(qml.PauliZ(0))]

        # Raises an error if queue or observables are invalid
        mock_device_supporting_paulis.check_validity(queue, observables)

    def test_check_validity_on_tensor_support(self, mock_device_supporting_paulis):
        """Tests the function Device.check_validity with tensor support capability"""
        queue = [
            qml.PauliX(wires=0),
            qml.PauliY(wires=1),
            qml.PauliZ(wires=2),
        ]

        observables = [qml.expval(qml.PauliZ(0) @ qml.PauliX(1))]

        # mock device does not support Tensor product
        with pytest.raises(DeviceError, match="Tensor observables not supported"):
            mock_device_supporting_paulis.check_validity(queue, observables)

    def test_check_validity_on_invalid_obserable_with_tensor_support(self, monkeypatch):
        """Tests the function Device.check_validity with tensor support capability
        but with an invalid observable"""
        queue = [
            qml.PauliX(wires=0),
            qml.PauliY(wires=1),
            qml.PauliZ(wires=2),
        ]

        observables = [qml.expval(qml.PauliZ(0) @ qml.Hadamard(1))]

        D = Device
        with monkeypatch.context() as m:
            m.setattr(D, '__abstractmethods__', frozenset())
            m.setattr(D, 'operations', ["PauliX", "PauliY", "PauliZ"])
            m.setattr(D, 'observables', ["PauliX", "PauliY", "PauliZ"])
            m.setattr(D, 'capabilities', lambda self: {"tensor_observables": True})
            m.setattr(D, 'short_name', "Dummy")

            dev = D()

            # mock device supports Tensor products but not hadamard
            with pytest.raises(DeviceError, match="Observable Hadamard not supported"):
                dev.check_validity(queue, observables)

    def test_check_validity_on_invalid_queue(self, mock_device_supporting_paulis):
        """Tests the function Device.check_validity with invalid queue and valid observables"""
        queue = [
            qml.RX(1.0, wires=0),
            qml.PauliY(wires=1),
            qml.PauliZ(wires=2),
        ]

        observables = [qml.expval(qml.PauliZ(0))]

        with pytest.raises(DeviceError, match="Gate RX not supported on device"):
            mock_device_supporting_paulis.check_validity(queue, observables)

    def test_check_validity_on_invalid_observable(self, mock_device_supporting_paulis):
        """Tests the function Device.check_validity with valid queue and invalid observables"""
        queue = [
            qml.PauliX(wires=0),
            qml.PauliY(wires=1),
            qml.PauliZ(wires=2),
        ]

        observables = [qml.expval(qml.Hadamard(0))]

        with pytest.raises(DeviceError, match="Observable Hadamard not supported on device"):
            mock_device_supporting_paulis.check_validity(queue, observables)


class TestClassmethods:
    """Test the classmethods of Device"""

    def test_capabilities(self, mock_device_with_capabilities):
        """check that device can give a dict of further capabilities"""

        assert mock_device_with_capabilities.capabilities() == mock_device_capabilities


class TestOperations:
    """Tests the logic related to operations"""

    def test_shots_setter(self, mock_device):
        """Tests that the property setter of shots changes the number of shots."""

        assert mock_device._shots == 1000

        mock_device.shots = 10

        assert mock_device._shots == 10

    @pytest.mark.parametrize("shots", [-10, 0])
    def test_shots_setter_error(self, mock_device, shots):
        """Tests that the property setter of shots raises an error if the requested number of shots
        is erroneous."""

        with pytest.raises(qml.DeviceError, match="The specified number of shots needs to be at least 1"):
            mock_device.shots = shots

    def test_op_queue_accessed_outside_execution_context(self, mock_device):
        """Tests that a call to op_queue outside the execution context raises the correct error"""

        with pytest.raises(
            ValueError, match="Cannot access the operation queue outside of the execution context!"
        ):
            mock_device.op_queue

    def test_op_queue_is_filled_at_pre_measure(self, mock_device_with_paulis_and_methods, monkeypatch):
        """Tests that the op_queue is correctly filled when pre_measure is called and that accessing
           op_queue raises no error"""
        queue = [
            qml.PauliX(wires=0),
            qml.PauliY(wires=1),
            qml.PauliZ(wires=2),
        ]

        observables = [
            qml.expval(qml.PauliZ(0)),
            qml.var(qml.PauliZ(1)),
            qml.sample(qml.PauliZ(2)),
        ]

        queue_at_pre_measure = []

        with monkeypatch.context() as m:
            m.setattr(Device, 'pre_measure', lambda self: queue_at_pre_measure.extend(self.op_queue))
            mock_device_with_paulis_and_methods.execute(queue, observables)

        assert queue_at_pre_measure == queue

    def test_op_queue_is_filled_during_execution(self, mock_device_with_paulis_and_methods, monkeypatch):
        """Tests that the operations are properly applied and queued"""
        queue = [
            qml.PauliX(wires=0),
            qml.PauliY(wires=1),
            qml.PauliZ(wires=2),
        ]

        observables = [
            qml.expval(qml.PauliZ(0)),
            qml.var(qml.PauliZ(1)),
            qml.sample(qml.PauliZ(2)),
        ]

        call_history = []
        with monkeypatch.context() as m:
            m.setattr(Device, 'apply', lambda self, op, wires, params: call_history.append([op, wires, params]))
            mock_device_with_paulis_and_methods.execute(queue, observables)

        assert call_history[0] == ["PauliX", [0], []]
        assert call_history[1] == ["PauliY", [1], []]
        assert call_history[2] == ["PauliZ", [2], []]

    def test_unsupported_operations_raise_error(self, mock_device_with_paulis_and_methods):
        """Tests that the operations are properly applied and queued"""
        queue = [
            qml.PauliX(wires=0),
            qml.PauliY(wires=1),
            qml.Hadamard(wires=2),
        ]

        observables = [
            qml.expval(qml.PauliZ(0)),
            qml.var(qml.PauliZ(1)),
            qml.sample(qml.PauliZ(2)),
        ]

        with pytest.raises(DeviceError, match="Gate Hadamard not supported on device"):
            mock_device_with_paulis_and_methods.execute(queue, observables)


class TestObservables:
    """Tests the logic related to observables"""

    def test_obs_queue_accessed_outside_execution_context(self, mock_device):
        """Tests that a call to op_queue outside the execution context raises the correct error"""

        with pytest.raises(
            ValueError,
            match="Cannot access the observable value queue outside of the execution context!",
        ):
            mock_device.obs_queue

    def test_obs_queue_is_filled_at_pre_measure(self, mock_device_with_paulis_and_methods, monkeypatch):
        """Tests that the op_queue is correctly filled when pre_measure is called and that accessing
           op_queue raises no error"""
        queue = [
            qml.PauliX(wires=0),
            qml.PauliY(wires=1),
            qml.PauliZ(wires=2),
        ]

        observables = [
            qml.expval(qml.PauliZ(0)),
            qml.var(qml.PauliZ(1)),
            qml.sample(qml.PauliZ(2)),
        ]

        queue_at_pre_measure = []

        with monkeypatch.context() as m:
            m.setattr(Device, 'pre_measure', lambda self: queue_at_pre_measure.extend(self.obs_queue))
            mock_device_with_paulis_and_methods.execute(queue, observables)

        assert queue_at_pre_measure == observables

    def test_obs_queue_is_filled_during_execution(self, monkeypatch, mock_device_with_paulis_and_methods):
        """Tests that the operations are properly applied and queued"""
        observables = [
            qml.expval(qml.PauliX(0)),
            qml.var(qml.PauliY(1)),
            qml.sample(qml.PauliZ(2)),
        ]

        # capture the arguments passed to dev methods
        expval_args = []
        var_args = []
        sample_args = []
        with monkeypatch.context() as m:
            m.setattr(Device, 'expval', lambda self, *args: expval_args.extend(args))
            m.setattr(Device, 'var',    lambda self, *args: var_args.extend(args))
            m.setattr(Device, 'sample', lambda self, *args: sample_args.extend(args))
            mock_device_with_paulis_and_methods.execute([], observables)

        assert expval_args == ["PauliX", [0], []]
        assert var_args == ["PauliY", [1], []]
        assert sample_args == ["PauliZ", [2], []]


    def test_unsupported_observables_raise_error(self, mock_device_with_paulis_and_methods):
        """Tests that the operations are properly applied and queued"""
        queue = [
            qml.PauliX(wires=0),
            qml.PauliY(wires=1),
            qml.PauliZ(wires=2),
        ]

        observables = [
            qml.expval(qml.Hadamard(0)),
            qml.var(qml.PauliZ(1)),
            qml.sample(qml.PauliZ(2)),
        ]

        with pytest.raises(DeviceError, match="Observable Hadamard not supported on device"):
            mock_device_with_paulis_and_methods.execute(queue, observables)

    def test_unsupported_observable_return_type_raise_error(self, mock_device_with_paulis_and_methods):
        """Check that an error is raised if the return type of an observable is unsupported"""

        queue = [qml.PauliX(wires=0)]

        # Make a observable without specifying a return operation upon measuring
        obs = qml.PauliZ(0)
        obs.return_type = "SomeUnsupportedReturnType"
        observables = [obs]

        with pytest.raises(QuantumFunctionError, match="Unsupported return type specified for observable"):
            mock_device_with_paulis_and_methods.execute(queue, observables)



class TestParameters:
    """Test for checking device parameter mappings"""

    def test_parameters_accessed_outside_execution_context(self, mock_device):
        """Tests that a call to parameters outside the execution context raises the correct error"""

        with pytest.raises(
            ValueError,
            match="Cannot access the free parameter mapping outside of the execution context!",
        ):
            mock_device.parameters

    def test_parameters_available_at_pre_measure(self, mock_device, monkeypatch):
        """Tests that the parameter mapping is available when pre_measure is called and that accessing
           Device.parameters raises no error"""


        p0 = 0.54
        p1 = -0.32

        queue = [
            qml.RX(p0, wires=0),
            qml.PauliY(wires=1),
            qml.Rot(0.432, 0.123, p1, wires=2),
        ]

        parameters = {0: (0, 0), 1: (2, 3)}

        observables = [
            qml.expval(qml.PauliZ(0)),
            qml.var(qml.PauliZ(1)),
            qml.sample(qml.PauliZ(2)),
        ]

        p_mapping = {}

        with monkeypatch.context() as m:
            m.setattr(Device, "pre_measure", lambda self: p_mapping.update(self.parameters))
            mock_device.execute(queue, observables, parameters=parameters)

        assert p_mapping == parameters


class TestDeviceInit:
    """Tests for device loader in __init__.py"""

    def test_no_device(self):
        """Test that an exception is raised for a device that doesn't exist"""

        with pytest.raises(DeviceError, match="Device does not exist"):
            qml.device("None", wires=0)

    def test_outdated_API(self, monkeypatch):
        """Test that an exception is raised if plugin that targets an old API is loaded"""

        with monkeypatch.context() as m:
            m.setattr(qml, "version", lambda: "0.0.1")
            with pytest.raises(DeviceError, match="plugin requires PennyLane versions"):
                qml.device("default.qubit", wires=0)
